"""Data structure, macros, and page generation for a glossary page (term dictionary and cross-reference)"""
import typing as tg

import base as b
import cache
import sdrl.constants as c
import sdrl.elements as el
import sdrl.macros as macros
import sdrl.markdown as md
import sdrl.partbuilder


class Glossary(sdrl.partbuilder.PartbuilderMixin, el.Part):
    """
    Processed in two phases: in phase 1, term references and term definitions are collected.
    Links to term definitions can already be generated because they have a canonical form:
    An anchor will exist on the glossary page for each alias of a term.
    In phase 2, the collected data are used to generate the actual glossary page or error messages.
    Defines macros TERM0 and TERM to be used in the glossary file and
    TERMREF to be used anywhere.
    """
    TEMPLATENAME = 'glossary.html'
    body_s: el.Body_s  # we derive body_i 
    course: 'sdrl.course.Coursebuilder'
    explainedby: dict[str, set[str]]  # explainedby[term] == partnames_with_explain
    mentionedby: dict[str, set[str]]  # mentionedby[term] == partnames_with_termref
    termdefs: set[str]  # what has a [TERMx] in glossary.md
    term_linkslist: tg.Optional[list[str]]  # the lines of the linksblock
    rendered_content: str = ''  # results of render(); re-render would report duplicate definitions
    includefiles: list[str]  # files INCLUDEd while rendering the glossary
    
    def __init__(self, name, **kwargs):
        super().__init__(name, **kwargs)
        self.read_partsfile(self.sourcefile)
        b.copyattrs(self.sourcefile, self.metadata, self,
                    mustcopy_attrs='title',
                    cancopy_attrs='stage',
                    mustexist_attrs='')
        self.explainedby = dict()
        self.mentionedby = dict()
        self.termdefs = set()
        self.term_linkslist = None  # set by [TERM], used and unset by [ENDTERM]
        self.register_macros_phase1()
        self.directory.make_the(el.Sourcefile, self.sourcefile)
        self.make_dependency(el.Topmatter, part=self)
        self.make_dependency(el.Glossarybody, part=self, includelist_class=el.IncludeList_s,
                             switch_macros_op=self.register_macros_phase2)
        self.make_dependency(el.Toc, part=self)

    def check_existing_resource(self):
        super().check_existing_resource()
        if self.state == cache.State.AS_BEFORE and self.cache.is_dirty(self.sourcefile):
            self.state = cache.State.HAS_CHANGED  # ensure repetition of error messages

    def do_build(self):
        b.debug(f"do_build({self.cache_key}):")
        body = self.directory.get_the(el.Glossarybody, self.name).value  # noqa
        self.render_structure(self.course, self, body, self.targetdir_s)  # noqa
        self.render_structure(self.course, self, body, self.targetdir_i, info=False)  # noqa
        self.report_issues()

    @property
    def sourcefile(self) -> str:
        return f"{self.course.chapterdir}/{c.AUTHOR_GLOSSARY_BASENAME}.md"

    @property
    def toc(self) -> str:
        """Return a chapters-only table of contents for the glossary."""
        result = ['']  # start with a newline
        for chapter in self.course.chapters:  # noqa
            if chapter.to_be_skipped:
                continue
            result.append(chapter.toc_entry)
        result.append(self.toc_entry)
        return "\n".join(result)

    @property
    def toc_link_text(self) -> str:
        return f"<a href='{self.outputfile}'>{self.title}</a>"

    def explains(self, partname: str, term: str):  # called by author.py
        if term not in self.explainedby:
            self.explainedby[term] = set()
        if not self.course.get_part(self.sourcefile, partname).to_be_skipped:
            self.explainedby[term].add(partname)

    def mentions(self, partname: str, term: str):
        if term not in self.mentionedby:
            self.mentionedby[term] = set()
        if not self.course.get_part(self.sourcefile, partname).to_be_skipped:
            self.mentionedby[term].add(partname)

    def fill_mentionedbylists(self):
        """
        explainedby is filled because of Part.process_topmatter().
        mentionedby is incompletely filled because most markdown does not get rendered in a given build.
        We fill it here properly from the TermrefList objects.
        """
        for termreflist in self.directory.get_all(el.TermrefList):
            for term in termreflist.value:
                self.mentions(termreflist.part.name, term)

    def render(self, mode: b.Mode):
        """We render only once, because the glossary will not contain [INSTRUCTOR] calls."""
        if not self.rendered_content:
            self.register_macros_phase2()
            mddict = md.render_markdown(self.sourcefile, self.name, self.content, mode, dict())
            self.rendered_content, self.includefiles = (mddict['html'], mddict['includefiles'])
    
    def report_issues(self):
        terms_explained = set(self.explainedby.keys())
        terms_mentioned = set(self.mentionedby.keys())
        undefined_terms = (terms_explained | terms_mentioned) - self.termdefs
        if undefined_terms:
            what = "This term lacks" if len(undefined_terms) == 1 else "These terms lack"
            b.error(f"{what} a definition: {sorted(undefined_terms)}", file=self.sourcefile)

    def register_macros_phase1(self):
        macros.register_macro("TERMREF", 1, macros.MM.INNER, self._expand_termref)
        macros.register_macro("TERMREF2", 2, macros.MM.INNER, self._expand_termref)
        # --- the following will be redefined in phase2:
        macros.register_macro("TERM0", 1, macros.MM.INNER, self._complain_term)
        macros.register_macro("TERM", 1, macros.MM.BLOCKSTART, self._complain_term)
        macros.register_macro("ENDTERM", 0, macros.MM.BLOCKEND, self._ignore_endtermlong)

    def register_macros_phase2(self):
        macros.register_macro("TERM0", 1, macros.MM.INNER, self._expand_term0, redefine=True)
        macros.register_macro("TERM", 1, macros.MM.BLOCKSTART, self._expand_term, redefine=True)
        macros.register_macro("ENDTERM", 0, macros.MM.BLOCKEND, self._expand_endterm, redefine=True)

    def _expand_termref(self, macrocall: macros.Macrocall) -> str:
        term = macrocall.arg1
        label = term if not macrocall.arg2 else macrocall.arg2  # unify calls to TERMREF and TERMREF2
        if label.startswith("-"):  # arg2 is meant to be a suffix
            label = f"{term}{label[1:]}"
        elif label.startswith("&ndash;"):  # e.g. [TERMREF2::class::--like] for 'class-like', but
            # Markdown extension smarty has garbled the '--' into '&ndash;' 
            label = f"{term}-{label[len('&ndash;'):]}"
        target = "%s.html#%s" % (c.AUTHOR_GLOSSARY_BASENAME, b.slugify(term))
        self._macro_mentions(macrocall, term)
        return (f"<a href='{target}' class='glossary-termref-term'>"
                f"{label}<span class='glossary-termref-suffix'></span></a>")

    def _complain_term(self, mc: macros.Macrocall) -> str:  # noqa
        b.error(f"{mc.macrocall_text}': [{mc.macroname}] can only be used in the glossary", file=mc.filename)
        return ""  # no expansion in phase 1

    def _ignore_endtermlong(self, macrocall: macros.Macrocall) -> str:  # noqa
        return ""

    def _expand_term0(self, macrocall: macros.Macrocall) -> str:
        """[TERM0::term|second form of term|third form|and so on], no definition text is supplied"""
        macrocall.arg2 = ""  # empty body
        return self._expand_termdef(macrocall)

    def _expand_term(self, macrocall: macros.Macrocall) -> str:
        """[TERM::term|second form of term|et cetera]"""
        open_body = "\n<div class='glossary-term-body'>\n\n"
        result = "".join(self._expand_any_termdef(macrocall)) + open_body
        return result
    
    def _expand_termdef(self, macrocall: macros.Macrocall) -> str:
        """allows two-argument macro calls that include a short definition"""
        termdef = macrocall.arg2
        result = self._expand_any_termdef(macrocall)
        # ----- generate body:
        if termdef:
            result.append(f"<span class='glossary-term-body'>{termdef}</span>\n\n")
        # ----- generate linkblock:
        result.extend(self.term_linkslist)
        self.term_linkslist = None
        # ----- done!:
        return "".join(result)

    def _expand_endterm(self, macrocall: macros.Macrocall) -> str:
        """[ENDTERM]  (a [TERM] has to be open)"""
        close_body = "\n\n</div>\n"
        if self.term_linkslist is not None:
            result = self.term_linkslist
            self.term_linkslist = None
            return close_body + "".join(result)
        else:
            b.error(f"[ENDTERM] is lacking its [TERM::...]", file=macrocall.filename)
            return ""

    @staticmethod
    def _collect_parts(termrefdict: dict[str, set[str]], termslist: list[str]) -> set[str]:
        """
        termslist is the list of aliases of a term.
        termrefdict maps a term (main or alias) to a set of partnames refering to it.
        Returns the set of partnames refering to one of the terms in termslist: The union
        of the respective termrefdict entry sets.
        """
        result = set()
        for term in termslist:
            if term in termrefdict:
                result |= termrefdict[term]
        return result

    def _expand_any_termdef(self, macrocall: macros.Macrocall) -> list[str]:  # [TERM0], [TERM]
        """Sets self.term_linkslist as a side effect."""
        separator = '|'
        myfile, mypart = macrocall.filename, macrocall.partname
        terms = macrocall.arg1
        termslist = terms.split(separator)
        headingtext = " | ".join(termslist)  # we keep the original order
        result = []
        # ----- report duplicate entries:
        for term in termslist:
            if term in self.termdefs:
                b.error(f"{macrocall.macrocall_text}: Term '{term}' is already defined", file=macrocall.filename)
            self.termdefs.add(term)
        # ----- open block:
        result.append("\n<div class='glossary-term-block'>\n")
        # ----- generate anchors:
        anchors = ("<a id='%s'></a>\n" % b.slugify(term) for term in termslist)
        result.extend(anchors)
        # ----- generate heading:
        result.append(f"<span class='glossary-term-heading'>{headingtext}</span>\n")
        # ----- generate links:
        links = []
        explainedby_names = self._collect_parts(self.explainedby, termslist)
        mentionedby_names = self._collect_parts(self.mentionedby, termslist)
        explainedby_links = sorted((f"[PARTREF::{p}]" for p in explainedby_names))
        mentionedby_links = sorted((f"[PARTREF::{p}]" for p in mentionedby_names))
        any_links = explainedby_links or mentionedby_links
        if any_links:
            links.append("\n<div class='glossary-term-linkblock'>\n")
        if explainedby_links:
            links.append(" <div class='glossary-term-links-explainedby'>\n   ")
            links.append(macros.expand_macros(myfile, mypart,  ", ".join(explainedby_links)))
            links.append("\n </div>\n")
        if mentionedby_links:
            links.append(" <div class='glossary-term-links-mentionedby'>\n")
            links.append("  " + macros.expand_macros(myfile, mypart, ", ".join(mentionedby_links)))
            links.append("\n </div>\n")
        if any_links:
            links.append("</div>\n")
        # ----- close block:
        result.append("\n</div>\n")
        if self.term_linkslist is not None:
            b.error(f"{macrocall.macrocall_text} is preceeded by a [TERM] with no [ENDTERM]", 
                    file=macrocall.filename)
        self.term_linkslist = links
        # ----- done!:
        return result

    def _macro_mentions(self, macrocall: macros.Macrocall, term: str):  # called by phase 1 TERMREF macro expansion
        partname = macrocall.partname
        if partname and partname != c.AUTHOR_GLOSSARY_BASENAME:  # avoid links from glossary to itself
            macrocall.md.termrefs.add(term)



"""Shortcut typenames, global constants, basic helpers."""
import enum
import json
import logging
import re
import time
import typing as tg

import requests
import rich
import rich.progress
import rich.table
import yaml


starttime = time.time()
num_errors = 0
msgs_seen = set()
loglevel = logging.ERROR
loglevels = dict(DEBUG=logging.DEBUG, INFO=logging.INFO, WARNING=logging.WARNING,
                 ERROR=logging.ERROR, CRITICAL=logging.CRITICAL)
register_files_callback: tg.Callable[[str], None]

OStr = tg.Optional[str]
StrAnyDict = dict[str, tg.Any]  # JSON or YAML structures


def set_loglevel(level: str):
    global loglevels, loglevel
    if level in loglevels:
        global loglevel
        loglevel = loglevels[level]
    else:
        pass  # simply ignore nonexisting loglevels


def set_register_files_callback(callback: tg.Callable[[str], None]):
    global register_files_callback
    register_files_callback = callback


class Mode(enum.Enum):
    STUDENT = "student"
    INSTRUCTOR = "instructor"


class CritialError(Exception):
    pass

def as_fingerprint(raw: str) -> str:
    """Canonicalize fingerprint: all-lowercase, no blanks"""
    return raw.replace(' ', '').lower()


def copyattrs(context: str, source: StrAnyDict, target: tg.Any, 
              mustcopy_attrs: str, cancopy_attrs: str, mustexist_attrs: str, 
              typecheck: dict[str, type]=dict(), overwrite=True, report_extra=False):  # noqa
    """
    Copies data from YAML or JSON mapping 'source' to class object 'target' and checks attribute set of d.
    mustcopy_attrs, cancopy_attrs, and mustexist_attrs are comma-separated attribute name lists.
    mustcopy_attrs and cancopy_attrs are copied; mustcopy_attrs and mustexist_attrs must exist; 
    cancopy_attrs need not exist.
    typecheck defines types for attrs that must not be str.
    If overwrite is False, fails if attribute already exists in target.
    Prints error and stops on problems, using 'context' as location info in the error message.
    E.g. copyattrs(sourcefile, yaml, self, "title,shorttitle,dir", "templatedir", "chapters")
    """
    def mysetattr(obj, name, newvalue):
        if overwrite or not hasattr(target, name):
            setattr(obj, name, newvalue)
    
    def names_in(attrlist: str) -> list[str]:
        if not attrlist:
            return []
        return [a.strip() for a in attrlist.split(',')]
    mustcopy_names = names_in(mustcopy_attrs)  # mandatory
    cancopy_names = names_in(cancopy_attrs)  # optional
    mustexist_names = names_in(mustexist_attrs)  # further mandatory
    if not source:
        source = dict()
    source_names = set(source.keys())
    for mname in mustcopy_names:  # transport these
        value = source.get(mname, ValueError)
        if value is ValueError:
            error(f"{context}: required attribute is missing: {mname}")
            finalmessage()
        else:
            mysetattr(target, mname, value)
    for cname in cancopy_names:  # transport these if present, set them to None if not
        if cname in source:
            if hasattr(target, cname) and not overwrite:
                warning("%s: %soverwriting '%s': old value '%s', new value '%s'" %
                        (context, "" if overwrite else "not ", cname, getattr(target, cname), source[cname]))
            mysetattr(target, cname, source[cname])
        elif cname not in source and not hasattr(target, cname):
            setattr(target, cname, None)
    extra_attrs = source_names - set(mustcopy_names) - set(cancopy_names) - set(mustexist_names)
    if report_extra and extra_attrs:
        warning(f"unexpected extra attributes found: {extra_attrs}", file=context)
    for attrname, its_type in typecheck.items():
        value = getattr(target, attrname, None)
        if value and not isinstance(value, its_type):
            error(f"'{context}': attribute '{attrname}' should be {str(its_type)} (is '{value}')")


def slurp(resource: str) -> str:
    """Reads local file (via filename) or http resource (via URL)."""
    try:
        if resource.startswith('http:') or resource.startswith('https://'):
            response = requests.get(resource)
            return response.text
        else:
            with open(resource, 'rt', encoding='utf8') as f:
                return f.read()
    except:  # noqa
        critical(f"'{resource}' does not exist")


def slurp_json(resource: str) -> StrAnyDict:
    return json.loads(slurp(resource))


def slurp_yaml(resource: str) -> StrAnyDict:
    return yaml.safe_load(slurp(resource))


def spit(filename: str, content: str):
    with open(filename, 'wt', encoding='utf8') as f:
        f.write(content)


def spit_json(filename: str, content: StrAnyDict):
    spit(filename, json.dumps(content))


def spit_yaml(filename: str, content: StrAnyDict):
    spit(filename, yaml.safe_dump(content))


def slugify(value: str) -> str:
    """ Slugify a string, to make it URL friendly. """
    separator = "-"
    value = re.sub(r'[^\w\s-]', '', value).strip().lower()
    return re.sub(r'[{}\s]+'.format(separator), separator, value)


def debug(msg: str):
    if loglevel <= logging.DEBUG:
        rich_print(msg)


def info(msg: str):
    if loglevel <= logging.INFO:
        rich_print(msg, "green")


def warning(msg: str, file: str = None, file2: str = None):
    if loglevel <= logging.WARNING:
        msg = _process_params(msg, file, file2)
        rich_print(msg, "yellow")


def error(msg: str, file: str = None, file2: str = None):
    if loglevel <= logging.ERROR:
        msg = _process_params(msg, file, file2)
        rich_print(msg, "red", count=1)


def critical(msg: str):
    rich_print(msg, "bold red", count=1)
    raise CritialError(msg)


def finalmessage():
    timing = "%.1f seconds" % (time.time() - starttime)
    if num_errors > 0:
        critical(f"==== {num_errors} error{'s' if num_errors != 1 else ''}. {timing}. Exiting. ====")
    else:
        info(f"::: {timing} :::")


def caller(how_far_up: int = 1) -> str:
    """Returns 'functionname:lineno' for how_far_up frames above the caller: the caller's n-th caller"""
    import inspect as sp
    frame = sp.currentframe().f_back  # the caller's frame
    for k in range(how_far_up):
        frame = frame.f_back
    return f"{frame.f_code.co_qualname}:{frame.f_lineno}"


def plural_s(number, value="s") -> str:
    return value if number != 1 else ""


def Table() -> rich.table.Table:
    """An empty Table in default sedrila style"""
    return rich.table.Table(show_header=True, header_style="bold yellow",
                            show_edge=False, show_footer=False)


def get_progressbar(maxcount: int) -> tg.Iterator[rich.progress.ProgressType]:
    return iter(rich.progress.track(range(maxcount), description="", transient=True, ))


def rich_print(msg: str, enclose_in_tag: tg.Optional[str] = None, count = 0):
    """Print any message, but each one only once."""
    global num_errors, msgs_seen
    if msg not in msgs_seen:
        msgs_seen.add(msg)
        num_errors += count
        if enclose_in_tag:
            msg = f"[{enclose_in_tag}]{msg}[/{enclose_in_tag}]"            
        rich.print(msg)


def _process_params(msg: str, file: tg.Optional[str], file2: tg.Optional[str]):
    if file and file2:
        msg = f"Files '{file}' and '{file2}':\n   {msg}"
        register_files_callback(file)
        register_files_callback(file2)
    elif file:
        msg = f"File '{file}':\n   {msg}"
        register_files_callback(file)
    return msg


def _testmode_reset():
    """reset error counter; avoid text wrapping of b.error() etc."""
    global num_errors, msgs_seen, starttime
    starttime = time.time()
    num_errors = 0
    msgs_seen = set()
    rich.get_console()._width = 10000
    set_register_files_callback(lambda s: None)


"""
Manage the replacements data for the <replacement id='abcde'>...</replacement> markdown extension
It comes from a file that contains as many such <replacement> groups as needed,
separated by whatever, e.g. two empty lines.
"""
import re

import base as b

replacement_expr_re = r"<replacement +?id=[\'\"]([\w_-]+?)[\'\"]>(.+?)</replacement>"
replacementsdict = dict()
replacements_loaded = False  # nothing has been loaded, so just keep the original content


def load_replacements_file(filename: str):
    with open(filename, 'rt', encoding='utf8') as f:
        load_replacements_string(filename, f.read())


def load_replacements_string(context_filename: str, s: str):
    global replacementsdict, replacements_loaded
    for repl_id, body in re.findall(replacement_expr_re, s, flags=re.DOTALL):
        if repl_id in replacementsdict:
            b.warning(f"replacement '{repl_id}' is defined multiple times", file=context_filename)
        replacementsdict[repl_id] = body
    replacements_loaded = True


def get_replacement(context_filename: str, content: str, repl_id: str) -> str:
    global replacementsdict, replacements_loaded
    if not replacements_loaded:
        return content
    if repl_id not in replacementsdict:
        b.warning(f"replacement '{repl_id}' is not defined, using '????'", file=context_filename)
    return replacementsdict.get(repl_id, "????")



import dataclasses
import functools
import os
import re

import base as b
import html
import sdrl.course
import sdrl.macros as macros
import sdrl.markdown as md

ALTDIR_KEYWORD = "ALT:"


def register_macros(course: sdrl.course.Coursebuilder):
    MM = macros.MM
    b.debug("registering macros")
    # ----- register EARLY-mode macros:
    macros.register_macro('INCLUDE', 1, MM.EARLY,
                          functools.partial(expand_include, course))
    # ----- register INNER-mode macros:
    macros.register_macro('HREF', 1, MM.INNER,
                          functools.partial(expand_href, course))  # show and link a URL
    macros.register_macro('PARTREF', 1, MM.INNER, 
                          functools.partial(expand_partref, course))  # name as linktext
    macros.register_macro('PARTREFTITLE', 1, MM.INNER, 
                          functools.partial(expand_partref, course))  # title as linktext
    macros.register_macro('PARTREFMANUAL', 2, MM.INNER, 
                          functools.partial(expand_partref, course))  # explicit linktext
    macros.register_macro('TREEREF', 1, MM.INNER, 
                          functools.partial(expand_treeref, course))  # show full path in itree
    macros.register_macro('EC', 0, MM.INNER, expand_enumeration, partswitch_enumeration)
    macros.register_macro('EQ', 0, MM.INNER, expand_enumeration, partswitch_enumeration)
    macros.register_macro('ER', 0, MM.INNER, expand_enumeration, partswitch_enumeration)
    macros.register_macro('EREFC', 1, MM.INNER, expand_enumerationref)
    macros.register_macro('EREFQ', 1, MM.INNER, expand_enumerationref)
    macros.register_macro('EREFR', 1, MM.INNER, expand_enumerationref)
    macros.register_macro('DIFF', 1, MM.INNER, sdrl.course.Taskbuilder.expand_diff)
    # ----- register hard-coded block macros:
    macros.register_macro('PROT', 1, MM.BLOCK, functools.partial(expand_prot, course))
    macros.register_macro('SECTION', 2, MM.BLOCKSTART, expand_section)
    macros.register_macro('ENDSECTION', 0, MM.BLOCKEND, expand_section)
    macros.register_macro('INNERSECTION', 2, MM.BLOCKSTART, expand_section)
    macros.register_macro('ENDINNERSECTION', 0, MM.BLOCKEND, expand_section)
    macros.register_macro('HINT', 1, MM.BLOCKSTART, expand_hint)
    macros.register_macro('ENDHINT', 0, MM.BLOCKEND, expand_hint)
    macros.register_macro('FOLDOUT', 1, MM.BLOCKSTART, expand_foldout)
    macros.register_macro('ENDFOLDOUT', 0, MM.BLOCKEND, expand_foldout)
    # ----- register generic block macros:
    generic_defs = {key: val for key, val in course.blockmacro_topmatter.items() if key.isupper()}
    for key, value in generic_defs.items():
        if "{arg2}" in value:
            args = 2
        elif "{arg1}" in value:
            args = 1
        else:
            args = 0
        macros.register_macro(key, args, MM.BLOCKSTART, expand_block)
        macros.register_macro(f'END{key}', 0, MM.BLOCKEND, expand_block)


def expand_href(course: sdrl.course.Coursebuilder, macrocall: macros.Macrocall) -> str:  # noqa
    return f"<a href='{macrocall.arg1}'>{macrocall.arg1}</a>"


def expand_partref(course: sdrl.course.Coursebuilder, macrocall: macros.Macrocall) -> str:
    part = course.get_part(macrocall.filename, macrocall.arg1)
    linktext = dict(PARTREF=part.name, 
                    PARTREFTITLE=part.title, 
                    PARTREFMANUAL=macrocall.arg2)[macrocall.macroname]
    return f"<a href='{part.outputfile}' class='partref-link'>{html.escape(linktext)}</a>"


def expand_treeref(course: sdrl.course.Coursebuilder, macrocall: macros.Macrocall) -> str:
    actualpath = includefile_path(course, macrocall, itree_mode=True)
    showpath = actualpath[len(course.itreedir)+1:]  # skip itreedir part of path
    if not os.path.exists(actualpath):
        b.warning(f"{macrocall.macrocall_text}: itreedir file '{actualpath}' not found",
                  file=macrocall.filename)
        showpath = "???"
    prefix = "<span class='treeref-prefix'></span>"
    mainpart = f"<span class='treeref'>{html.escape(showpath, quote=False)}</span>"
    suffix = "<span class='treeref-suffix'></span>"
    return f"{prefix}{mainpart}{suffix}"


def expand_prot(course: sdrl.course.Course, macrocall: macros.Macrocall) -> str:
    """[PROT::somedir/file.prot]. Plain paths in viewer mode, INCLUDE-style paths in author mode."""
    path = macrocall.arg1
    author_mode = isinstance(course, sdrl.course.Coursebuilder)  # in viewer mode we receive a dummy
    if author_mode:
        assert isinstance(course, sdrl.course.Coursebuilder)
        path = includefile_path(course, macrocall, itree_mode=False)
    if not os.path.exists(path):
        b.warning(f"{macrocall.macrocall_text}: file '{path}' not found", file=macrocall.filename)
        return f"\n<p>(('{path}' not found))</p>\n"

    @dataclasses.dataclass
    class State:
        s: int
        promptcount: int

    def promptmatch(the_line: str) -> re.Match:
        r"""
        Whether line is a shell prompt as prescribed by the SeDriLa course rules.
        Canonical prompt:  export PS1="\u@\h \w \t \!\n\$ "
        This is a two-line prompt. This routine considers the first line only.
        Prompts are copy-pasted as plaintext, so possible ANSI sequences for color are no longer present.
        The group names front, userhost, dir, time, num, back are part of the function interface 
        (i.e. the caller knows they all exist in any match and can use the group contents).
        """
        front_re = r"(?P<front>^.*?)"  # any stuff up front, e.g. '(myvenv)' 
        userhost_re = r"(?P<userhost>[-\+\w]+@[-\+\w]+)"  # user@host
        sep_re = r"\s+"
        dir_re = r"(?P<dir>[/~]\S*)"  # anything whitespace-less starting with '~' or '/'
        time_re = r"(?P<time>\d\d:\d\d:\d\d)"  # e.g. '14:03:59'
        num_re = r"(?P<num>\d+)"
        back_re = r"(?P<back>.*$)"  # any stuff in the back
        prompt_re = f"{front_re}{userhost_re}{sep_re}{dir_re}{sep_re}{time_re}{sep_re}{num_re}{back_re}"
        return re.fullmatch(prompt_re, the_line)

    def handle_promptmatch():  # uses mm, result, state. Corresponds to promptmatch().
        state.promptcount += 1
        state.s = PROMPTSEEN
        promptindex = f"<span class='vwr-promptidx'>{state.promptcount}.</span>"
        front = f"<span class='vwr-front'>{esc('front')}</span>"
        userhost = f"<span class='vwr-userhost'>{esc('userhost')}</span>"
        dir = f"<span class='vwr-dir'>{esc('dir')}</span>"
        time = f"<span class='vwr-time'>{esc('time')}</span>"
        num = f"<span class='vwr-num'> {esc('num')} </span>"
        back = f"<span class='vwr-back'>{esc('back')}</span>"
        result.append(f"<tr><td>{promptindex} {front} {userhost} {dir} {time} {num} {back}</td></tr>")

    def esc(groupname: str) -> str:  # abbrev; uses mm
        return html.escape(mm.group(groupname))  # TODO_1_prechelt: make whitespace (indentation etc.) work

    # ----- expand_prot(): process lines:
    content = b.slurp(path)
    macrocall.md.includefiles.add(path)  # record that we have included this file
    result = ["\n<table class='vwr-table'>"]
    PROMPTSEEN, OUTPUT = (1, 2)
    state = State(s=OUTPUT, promptcount = 0)
    for line in content.split('\n'):
        line = line.rstrip()  # get rid of newline
        mm = promptmatch(line)
        if mm:
            handle_promptmatch()
        elif state.s == PROMPTSEEN:  # this is the command line
            result.append(f"<tr><td><span class='vwr-cmd'>{html.escape(line)}</span></td></tr>")
            state.s = OUTPUT
        elif state.s == OUTPUT:
            result.append(f"<tr><td><span class='vwr-output'>{html.escape(line)}</span></td></tr>")
        else:
            assert False
    result.append("</table>\n\n")
    return '\n'.join(result)


def expand_section(macrocall: macros.Macrocall) -> str:
    """
    [SECTION::goal::subtype1,subtype2] etc.
    [INNERSECTION] is equivalent and serves for one level of proper nesting if needed.
    (Nesting [SECTION][ENDSECTION] blocks happens to work, but for the wrong reasons.)
    A section [SECTION]...[ENDSECTION] like the ahove gets rendered into the following structure:
    <div class='section section-goal'>
      <div class='section-subtypes section-goal-subtypes'>
        <div class='section-subtype section-goal-subtype1'>section_goal_subtype1_topmatter</div>
        <div class='section-subtype section-goal-subtype2'>section_goal_subtype2_topmatter</div>
      </div>
      section_goal_topmatter
      the entire body of the section block
    </div>
    """
    sectiontype = macrocall.arg1
    sectionsubtypes = macrocall.arg2
    if macrocall.macroname in ('SECTION', 'INNERSECTION'):
        r = []  # noqa, results list, to be join'ed
        r.append(f"<section class='section section-{sectiontype}'>")  # level 1
        r.append(f"<div class='section-subtypes section-{sectiontype}-subtypes'>")  # level 2
        subtypeslist = sectionsubtypes.split(",")
        for subtype in subtypeslist:
            matter = topmatter(macrocall, f'section_{sectiontype}_{subtype}')
            r.append(f"<div class='section-subtype section-{sectiontype}-{subtype}'>{matter}</div>")
        r.append("</div>")  # end level 2
        r.append(topmatter(macrocall, f'section_{sectiontype}'))
        return "\n".join(r)
    elif macrocall.macroname in ('ENDSECTION', 'ENDINNERSECTION'):
        return "</section>"
    assert False, macrocall  # impossible


def expand_hint(macrocall: macros.Macrocall) -> str:
    """[HINT::description what it is about]"""
    if macrocall.macroname == 'HINT':
        content = topmatter(macrocall, 'hint').format(arg1=macrocall.arg1)
        return f"<details class='blockmacro blockmacro-hint'><summary>\n{content}\n</summary>\n"
    elif macrocall.macroname == 'ENDHINT':
        return "</details>"
    assert False, macrocall  # impossible


def expand_foldout(macrocall: macros.Macrocall) -> str:
    """[FOLDOUT]"""
    if macrocall.macroname == 'FOLDOUT':
        return f"<details class='blockmacro blockmacro-foldout'><summary>\n{macrocall.arg1}\n</summary>\n"
    elif macrocall.macroname == 'ENDFOLDOUT':
        return "</details>"
    assert False, macrocall  # impossible


def expand_block(macrocall: macros.Macrocall) -> str:
    is_end = macrocall.macroname.startswith('END')
    macroname = (macrocall.macroname if not is_end else macrocall.macroname[3:])
    if is_end:
        return f"</div>"
    else:
        content = topmatter(macrocall, macroname).format(arg1=macrocall.arg1, arg2=macrocall.arg2)
        return f"<div class='blockmacro blockmacro-{macroname.lower()}'>\n{content}"


def expand_enumeration(macrocall: macros.Macrocall) -> str:
    """[EC], [EQ], [ER]"""
    macroname = macrocall.macroname
    classname = f"enumeration-{macroname.lower()}"
    value = macros.get_state(macroname) + 1
    macros.set_state(macroname, value)
    return f"<span class='{classname}'>{value}</span>"


def partswitch_enumeration(macroname: str, newpartname: str):  # noqa
    macros.set_state(macroname, 0)  # is independent of newpartname


def expand_enumerationref(macrocall: macros.Macrocall) -> str:
    """[ECREF], [EQREF], [ERREF]"""
    # markup matches that of expand_enumeration(), argument is not checked in any way
    macroname = macrocall.macroname
    classname = f"enumeration-{macroname.lower()}"
    value = macrocall.arg1
    return f"<span class='{classname}'>{value}</span>"


def expand_include(course: sdrl.course.Coursebuilder, macrocall: macros.Macrocall) -> str:
    """
    [INCLUDE::filename] inserts file contents into the Markdown text.
    If the file has suffix *.md, it is macro-expanded beforehands,
    contents of all other files are inserted verbatim.
    A relative filename is relative to the location of the file containing the macro call.
    An absolute filename is relative to course.chapterdir.
    The format [INCLUDE::ALT:filename] refers to files in course.altdir,
    for relative names, the path from chapterdir to filename's dir is used below altdir.
    In the special case [INCLUDE::ALT:] with no filename at all, the filename is reused as well.
    """
    fullfilename = includefile_path(course, macrocall)
    # print(f"## fullfilename: {fullfilename} ({macrocall.filename})")
    if not os.path.exists(fullfilename):
        msgfunc = macrocall.warning if macrocall.arg1.startswith(ALTDIR_KEYWORD) else macrocall.error
        msgfunc(f"file '{fullfilename}' does not exist")  # noqa
        return ""
    with open(fullfilename, "rt", encoding='utf8') as f:
        rawcontent = f.read()
    macrocall.md.includefiles.add(fullfilename)  # record that we have included this file
    if fullfilename.endswith('.md'):
        return macros.expand_macros(md.md.context_sourcefile, md.md.partname, rawcontent)
    elif fullfilename.endswith('.prot'):
        macrocall.error("filename must be *.prot. Call ignored.")
        return ""  # ignore the entire macrocall
    else:
        return rawcontent


def includefile_path(course: sdrl.course.Coursebuilder, macrocall: macros.Macrocall, itree_mode=False) -> str:
    """
    Normal mode constructs normal paths in chapterdir and those with prefix 'ALT:' in altdir.
    itree mode constructs normal paths in itreedir and warns about 'ALT:' prefix if present.
    """
    arg_re = r"(?P<alt>" + ALTDIR_KEYWORD + r")?(?P<slash>/)?(?P<incfile>.*)"
    mm = re.fullmatch(arg_re, macrocall.arg1)
    is_alt = mm.group("alt") is not None
    if is_alt and itree_mode:
        b.warning(f"{macrocall.macrocall_text}: '{ALTDIR_KEYWORD}' prefix makes no sense here",
                  file=macrocall.filename)
        is_alt = False  # ignore the prefix
    is_abs = mm.group("slash") is not None
    inc_filepath = mm.group("incfile")  # e.g. ../chapterlevel_includefile
    ctx_filepath = macrocall.filename  # e.g. ch/chapter/group/task.md
    ctx_dirpath = os.path.dirname(ctx_filepath)  # e.g. ch/chapter/group
    ctx_basename = os.path.basename(ctx_filepath)  # e.g. task.md
    abs_topdir = (is_alt and course.altdir) or (itree_mode and course.itreedir) or course.chapterdir
    rel_dirpath = ctx_dirpath.replace(course.chapterdir, abs_topdir, 1)
    # print(f"## {macrocall.arg1} -> alt:{is_alt}, abs:{is_abs}, inc:{inc_filepath}, ctx:{ctx_dirpath}, "
    #       f"abs_topdir:{abs_topdir}, rel_dirpath:{rel_dirpath}")
    fullpath = os.path.join(abs_topdir if is_abs else rel_dirpath, inc_filepath or ctx_basename)
    return fullpath


def topmatter(macrocall: macros.Macrocall, name: str) -> str:
    topmatterdict = macrocall.md.blockmacro_topmatter  # noqa
    if name in topmatterdict:
        return topmatterdict[name]
    else:
        b.error(f"{macrocall.macrocall_text}: blockmacro_topmatter '{name}' is not defined in config",
                file=macrocall.filename)
        return ""  # neutral result



import argparse
import os
import typing as tg
import readline  # noqa, is active automatically for input()

import base as b
import git
import sdrl.constants as c
import sdrl.course
import sdrl.interactive as i
import sdrl.participant
import sdrl.repo as r


meaning = """Reports on course execution so far or prepares submission to instructor."""


def add_arguments(subparser):
    subparser.add_argument('--init', action='store_true',
                           help="start initialization for student repo directory")
    subparser.add_argument('--submission', action='store_true',
                           help=f"generate {c.SUBMISSION_FILE} with possible tasks to be checked by instructor")
    subparser.add_argument('--interactive', default="True", action=argparse.BooleanOptionalAction,
                           help="open interactive terminal interface to select tasks to submit")
    subparser.add_argument('--import-keys', action='store_true',
                           help="(re)import all public gpg keys for the given course")
    subparser.add_argument('--log', default="INFO", choices=b.loglevels.keys(),
                           help="Log level for logging to stdout (default: INFO)")


def execute(pargs: argparse.Namespace):
    b.set_loglevel(pargs.log)
    if pargs.init:
        init()
        return
    student = sdrl.participant.Student()
    course = sdrl.course.CourseSI(configdict=student.metadatadict, context=student.metadata_url)
    commits = git.commits_of_local_repo(reverse=True)
    r.compute_student_work_so_far(course, commits)
    entries, workhours_total, timevalue_total = r.student_work_so_far(course)
    if pargs.submission:
        entries = [entry for entry in entries if course.task(entry[0]).remaining_attempts]  # without final rejections
        prepare_submission_file(course, student.root, entries, pargs.interactive)
    elif pargs.import_keys:
        r.import_gpg_keys(course.instructors)
    else:
        report_student_work_so_far(course, entries, workhours_total, timevalue_total)


def init():
    course_url = os.path.dirname(input("Course URL: "))
    metadata = sdrl.participant.Student.get_metadata(course_url)
    init_data = metadata.get('init_data') or {}  # noqa
    prompts = sdrl.participant.Student.prompts(init_data.get('studentprompts') or {})
    data = dict(course_url=course_url)
    for value in prompts:
        data[value] = input(prompts[value] + ": ")
    b.spit_yaml(c.PARTICIPANT_FILE, data)
    b.info(f"wrote '{c.PARTICIPANT_FILE}'. Now commit and push it.")
    if not(metadata.get('instructors')):
        b.warning("No information about instructors present. Skipping key import.")
        return
    gpgimportwarning = init_data.get('gpgimportwarning') or """
        Next step is the import of the public keys from the instructors of the course.
        This is necessary to correctly show your progress.
        Press <Enter> to continue or Q <Enter> to abort:  """
    response = input(gpgimportwarning)
    if response and response in "Qq":
        b.critical("Abort.")
    r.import_gpg_keys(metadata['instructors'])


def prepare_submission_file(course: sdrl.course.Course, root: str, 
                            entries: tg.Sequence[r.ReportEntry], interactive: bool = False):
    if interactive:
        entries = i.select_entries(entries)
    if not entries:
        b.info("No entries to submit.")
        return
    # ----- write file:
    b.spit_yaml(os.path.join(root, c.SUBMISSION_FILE), r.submission_file_entries(entries))
    b.info(f"Wrote file '{c.SUBMISSION_FILE}'.")
    # ----- give instructions for next steps:
    b.info(f"1. Commit it with commit message '{c.SUBMISSION_COMMIT_MSG}'. Push it.")
    b.info(f"2. Then send the following to your instructor by email:")
    b.info(f"  Subject: Please check submission")
    b.info(f"  sedrila instructor {git.origin_remote_of_local_repo()}")
    show_instructors(course)


def report_student_work_so_far(course: sdrl.course.Course, entries: tg.Sequence[r.ReportEntry],
                               workhours_total: float, timevalue_total: float, out=None):
    b.info("Your work so far:")
    table = b.Table()
    table.add_column("Taskname")
    table.add_column("Workhours", justify="right")
    table.add_column("Timevalue", justify="right")
    table.add_column("reject/accept")
    for taskname, workhours, timevalue, rejections, accepted in entries:
        task = course.taskdict[taskname]
        ra_string = (c.INTERACT_ACCEPT_SYMBOL + " ") if task.is_accepted else ""
        if task.rejections > 0:
            ra_string += f"{c.INTERACT_REJECT_SYMBOL * task.rejections}"
            remaining = task.remaining_attempts
            allowed = task.allowed_attempts
            ra_string += f" ({remaining} of {allowed} remain)" if not task.is_accepted else ""
            if not remaining:
                ra_string = f"{c.SUBMISSION_REJECT_MARK} (after {allowed} attempt{b.plural_s(allowed)})"
        table.add_row(taskname, "%4.2f" % workhours, "%4.2f" % timevalue, ra_string)
        if out is not None:
            out.append((taskname, "%4.2f" % workhours, "%4.2f" % timevalue, ra_string))
    # table.add_section()
    table.add_row("[b]=TOTAL[/b]", "[b]%6.2f[/b]" % workhours_total, "[b]%6.2f[/b]" % timevalue_total, "")
    b.rich_print(table)


def show_instructors(course, with_gitaccount=False):
    b.info("The instructors for this course are:")
    for instructor in course.instructors:
        b.info(f"  {instructor['nameish']} <{instructor['email']}>")
        if with_gitaccount:
            b.info(f"     git account: {instructor['gitaccount']}")



"""
Class model for the stuff taking part in the incremental build.
Read the architecture description in docs/internal_notes.md first.

The items involved in the build process as inputs, outputs, intermediate products, 
or non-product pure-computation work steps are called Element.
They are arranged in the following inheritance hierarchy:

```
Element
  Product
    Piece
      Body
        Body_s
        Body_i
        Glossarybody
      Byproduct
        Content
        IncludeList_s
        IncludeList_i
        TermrefList
      FreshPiece
        LinkslistBottom
        Toc
      Topmatter
    Outputfile
      CopiedFile
      Part
        Course
          CourseSI
          Coursebuilder
        Chapter
          Chapterbuilder
        Taskgroup
          Taskgroupbuilder
        Task
          Taskbuilder
        Glossary
        Zipfile
  Source
    Configelement
    Sourcefile
    Zipdir
  Step
    MetadataDerivation
```


## How states drive the build

Any element can be `AS_BEFORE` or `HAS_CHANGED` or `MISSING` (the latter not for `Sources`).
These are defined in `cache.State`.
If the element is `MISSING`, it will be built, which turns the state into `HAS_CHANGED`.
If it `HAS_CHANGED`, it will be re-built.
If it is `AS_BEFORE`, it will be re-built only if a dependency `HAS_CHANGED`.


## Methods involved in the build

The above three state-related rules are implemented by the framework method `Element.build()`,
which relies on three Element-type-specific methods:  
- `check_existing_resource()` performs the state check by means of the cache.
- `my_dependencies()` returns the Elements this Element depends on.
- `do_build()` implements the actual build step for this particular type of element.


## How data are represented in the cache

`Steps` and `Outputfiles` need and have no cache representation.

`Pieces` have a `value` that is stored in the cache;
it is either a `str` or a `list[str]` or a `b.StrAnyDict`.

`Sources` are represented in the cache only by cache keys, with an empty value.
This serves to determine their state:
- cache key missing: `MISSING`.
- cache key written to cache in the present run: `HAS_CHANGED`.
- cache key is in persistent cache from a previous run: `AS_BEFORE` or `HAS_CHANGED`, 
  depending on the file's mtime (or file tree's youngest mtime).


## How data are represented in the build

`Elements` all have a number of attributes for management purposes, see the `Element` class.
Most are submitted to the constructor call, but for convenience, 
some of these are inherited via a `part` attribute.
In contrast to these purely organizational data, there is _substantial_ data that supports the
specific purpose of each type of element:

`Steps` are pure computation. They modify the data of `Parts`, but have no data representation themselves.

`CopiedFiles` are represented by their file only.

`Parts` are represented by a complex object as defined in `course.py`.
This object is created at initialization time, perhaps modified during a `Step`,
and turned into a file during `do_build()`, when that is called.

`Pieces` are represented by their `value` attribute as described above.


## Special Element types

All Element types in principle follow the build logic described under "methods involved" above.
However, some do it in an odd way:

`Byproducts` have no `do_build()` themselves and so never set their own `value` themselves.
Instead, their `value` is set by some other `Piece`'s `do_build()`.

`FreshPieces` do not use the cache to retrieve a previous value if they are `AS_BEFORE`.
Instead, they _always_ retrieve the previous value, _always_ re-build themselves, and then compare
the two in order to determine whether they are `AS_BEFORE` or `HAS_CHANGED`.
This means the build step happens in `check_existing_resource()` and `do_build()` only writes
a changed outcome to the cache.
"""

import os.path
import shutil
import typing as tg
import zipfile

import yaml

import base as b
import cache as c
import sdrl.directory as dir
import sdrl.macros as macros
import sdrl.markdown as md


class Element:  # abstract class
    """
    Common superclass of the stuff taking part in incremental build: Sources and Products.
    Directory defines an ordering of these classes A, B, C such that if any B has any A as a dependency,
    A will be before B in the ordering. This allows building Elements simply in order A, B, C, ...
    build() is mostly generic and consists of calls to the class-specific framework filler methods
    check_existing_resource(), my_dependencies() and do_build().
    """
    name: str  # path, filename, or partname
    part: 'Part'  # where to find inherited attrs, only set for non-Parts
    directory: dir.Directory  # inherited
    sourcefile: str  # path, inherited
    state: c.State
    dependencies: list['Element']

    def __init__(self, name: str, **kwargs):
        self.name = name
        self.dependencies = []
        for key, val in kwargs.items():
            setattr(self, key, val)  # store all keyword args in same-named attr
        self._inherit_from_part()

    def __repr__(self):
        return self.cache_key

    @property
    def cache(self) -> c.SedrilaCache:
        return self.directory.cache

    @property
    def cache_key(self) -> str:
        return f"{self.name}__{self.__class__.__name__.lower()}"  # e.g. MyPart__body_i

    @property
    def statelabel(self) -> str:
        return f"{self.state}{'*' if isinstance(self, Byproduct) else ''}"

    def build(self):
        """Generic framework operation."""
        def do_it():
            self.do_build()
            if self.state == c.State.MISSING:  # HAS_CHANGED and AS_BEFORE are acceptable
                self.state = c.State.HAS_CHANGED

        self.check_existing_resource()  # some do_build() rely on this to have happened
        if self.state != c.State.AS_BEFORE:
            b.debug(f"{self.__class__.__name__}.build({self.name}) local state:\t{self.statelabel} ")
            do_it()
            return
        for dep in self.my_dependencies():
            if dep.state != c.State.AS_BEFORE:
                which = f"{dep.__class__.__name__}({dep.name})"
                b.debug(f"{self.__class__.__name__}.build({self.name}) dependency {which} state:\t{dep.statelabel}")
                do_it()
                return
        # b.debug(f"{self.__class__.__name__}.build({self.name}) state:\t{self.statelabel}")

    def check_existing_resource(self):
        """
        Category-specific: To be implemented differently in Source, Piece, Outputfile.
        Set state==MISSING if no resource exists,
        set state==HAS_CHANGED if Source has changed (check may involve cache),
        set state==AS_BEFORE (and perhaps read Piece from cache) otherwise.
        """
        assert False, f"{self.__class__.__name__}.{self.name}.check_existing_resource() not defined"

    def my_dependencies(self) -> tg.Iterable['Element']:
        return self.dependencies

    def do_build(self):
        """Class-specific: perform actual build work."""
        assert False, f"{self.__class__.__name__}.do_build({self.name}) not defined"

    def add_dependency(self, dep: 'Element'):
        self.dependencies.append(dep)

    def make_dependency(self, mytype: type, **kwargs):
        name = kwargs.pop('name') if 'name' in kwargs else self.name
        elem = self.directory.make_the(mytype, name, **kwargs)
        self.add_dependency(elem)

    def make_or_get_dependency(self, mytype: type, **kwargs):
        name = kwargs.pop('name') if 'name' in kwargs else self.name
        elem = self.directory.make_or_get_the(mytype, name, **kwargs)
        self.add_dependency(elem)

    def cached_str(self) -> tuple[str, c.State]:
        return self.cache.cached_str(self.cache_key)

    def cached_list(self) -> tuple[list[str], c.State]:
        return self.cache.cached_list(self.cache_key)

    def cached_set(self) -> tuple[set[str], c.State]:
        return self.cache.cached_set(self.cache_key)

    def cached_dict(self) -> tuple[b.StrAnyDict, c.State]:
        return self.cache.cached_dict(self.cache_key)

    def _inherit_from_part(self):
        """
        Only non-Part Elements receive a part= argument and non-Part Elements exist only during 'author', 
        when directory and cache are used and sourcefiles are meaningful.
        """
        if getattr(self, 'part', None):  # inherit attrs, to make construction simpler:
            self.directory = self.part.directory
            self.sourcefile = self.part.sourcefile
            self.course = self.part.course


class Product(Element):  # abstract class
    """Abstract superclass for any kind of thing that gets built."""
    pass


class Piece(Product):  # abstract class
    """
    Abstract superclass for an internal outcome of build, managed in the cache or with help of the cache.
    Pieces have a value that is set by check_existing_resource() (if in the cache) or by do_build()
    or both.
    """
    CACHED_TYPE = 'str'  # which kind of value is in the cache
    SC = c.SedrilaCache  # abbrev
    READFUNC = dict(str=SC.cached_str, list=SC.cached_list, set=SC.cached_set, dict=SC.cached_dict)
    WRITEFUNC = dict(str=SC.write_str, list=SC.write_list, set=SC.write_set, dict=SC.write_dict)
    value: c.Cacheable  # build result, from Cache or from do_build()
    
    def check_existing_resource(self):
        value, self.state = self.READFUNC[self.CACHED_TYPE](self.cache, self.cache_key)
        if value is not None:
            self.value = value  # do not set it to None

    def handle_value_and_state(self, value):
        """
        Where Outputfile elements directly produce an effect during do_build(),
        Pieces only live in the cache. This method puts them there and sets value.
        However, a freshly built value is still AS_BEFORE if it is the same as in the cache.
        """
        self.value = value
        cache_value, cache_state = self.READFUNC[self.CACHED_TYPE](self.cache, self.cache_key)
        if cache_state != c.State.MISSING and self.value == cache_value:
            self.state = c.State.AS_BEFORE
        else: 
            self.state = c.State.HAS_CHANGED
            self.WRITEFUNC[self.CACHED_TYPE](self.cache, self.cache_key, self.value)
        b.debug(f"handle_value_and_state({self.cache_key}, c:{b.caller()}): "  # TODO 2: deactivate?
                f"cache: {cache_state},{'same' if self.value == cache_value else 'different'}_value"
                f" --> {self.state}")

    def has_value(self) -> bool:
        return hasattr(self, 'value')


class Body(Piece):  # abstract class
    includelist_class: type
    termrefs: set[str]

    def __init__(self, name: str, **kwargs):
        super().__init__(name, **kwargs)
        self.add_dependency(self.directory.make_or_get_the(Content, self.name, part=self))
        includelist = self.directory.make_the(self.includelist_class, self.name, part=self)
        self.add_dependency(includelist)
        includelist.check_existing_resource()
        if includelist.state == c.State.AS_BEFORE:  # a list is available
            for fname in includelist.value:
                self.make_or_get_dependency(Sourcefile, name=fname, posthoc=True)

    def check_existing_resource(self):
        super().check_existing_resource()
        if self.cache.is_dirty(self.sourcefile):
            self.state = c.State.HAS_CHANGED  # force re-build for previous dirty files

    def do_do_build(self, includelist_class: type, render_mode: b.Mode):
        # --- prepare:
        content = self.directory.get_the(Content, self.name)
        # --- build body and byproduct includeslist:
        # includeslist gets filled when building self, but is also a dependency of self!
        # As a dependency, it gets built earlier, so the includes found here will come into effect
        # only during the next run, via the cache.
        includeslist = self.directory.get_the(includelist_class, self.name)
        macros.switch_part(self.name)
        mddict = md.render_markdown(self.sourcefile, self.name, content.value, render_mode, 
                                    self.course.blockmacro_topmatter)
        html, includes, self.termrefs = (mddict['html'], mddict['includefiles'], mddict['termrefs'])
        self.handle_value_and_state(html)
        includeslist.handle_value_and_state(includes)


class Body_s(Body):
    """Student HTML page text content.  Byproducts: IncludeList_s, Termreflist."""
    def do_build(self):
        self.do_do_build(IncludeList_s, b.Mode.STUDENT)
        # --- build byproduct termreflist (body_i.termrefs ought to be identical to self.termrefs):
        termreflist = self.directory.get_the(TermrefList, self.name)
        termreflist.handle_value_and_state(self.termrefs)


class Body_i(Body):
    """Instructor HTML page text content.  Byproduct: IncludeList_i."""
    def do_build(self):
        self.do_do_build(IncludeList_i, b.Mode.INSTRUCTOR)


class Glossarybody(Body):
    """Glossarypage content.  Byproduct: IncludeList_s."""
    switch_macros_op: tg.Callable

    def __init__(self, name: str, **kwargs):
        super().__init__(name, **kwargs)
        # All Parts must have been created before, or this will do nothing:
        for termreflist in self.directory.get_all(TermrefList):
            self.add_dependency(termreflist)
        for topmatter in self.directory.get_all(Topmatter):
            self.add_dependency(topmatter)

    def do_build(self):
        self.switch_macros_op()        
        self.part.fill_mentionedbylists()  # noqa
        self.do_do_build(IncludeList_s, b.Mode.STUDENT)


class Byproduct(Piece):  # abstract class
    """
    A Byproduct gets built not by its own do_build() but by its main Product's,
    which must be built first and will set the Byproduct's 'value' attr and cache entry.
    check_existing_resource() retrieves the cache entry iff value is not set (when the
    main Product did not need to be built).
    In both cases, Byproduct.do_build() does nothing.
    Byproducts rely on their main Product's dependency checking; they have no dependenies of their own. 
    """
    def check_existing_resource(self):
        if self.has_value():  # if main product was built, all is done already
            pass 
        else:  # main product was not built, need to read cache
            super().check_existing_resource()
            assert self.has_value()  # if main product was not built, we cannot be MISSING

    def my_dependencies(self) -> list[Element]:
        return []

    def do_build(self):
        pass  # Byproducts get built by their corresponding main product calling handle_value_and_state()


class Content(Byproduct):
    """Markdown part of a Part sourcefile. Byproduct of Topmatter."""
    pass


class IncludeList_s(Byproduct):
    """List of names of files INCLUDEd by a Part in its student version."""
    CACHED_TYPE = 'set'  # which kind of value is in the cache


class IncludeList_i(Byproduct):
    """List of names of files INCLUDEd by a Part in its instructor version."""
    CACHED_TYPE = 'set'  # which kind of value is in the cache


class TermrefList(Byproduct):
    """
    List of names of terms TERMREF'd by Part or term self.name.
    Many of these will be empty, therefore TermrefList objects are created only if needed.
    They are implicit dependencies of the Glossary.
    """
    CACHED_TYPE = 'set'  # which kind of value is in the cache


class FreshPiece(Piece):
    """Piece of Task that has no dependency and is always built. The cache only determines whether it has changed."""
    FRESH_ATTR = '?'  # attr of task that represents the piece's value

    def check_existing_resource(self):
        cached_value, cached_state = self.READFUNC[self.CACHED_TYPE](self.cache, self.cache_key)
        self.value = getattr(self.part, self.FRESH_ATTR)  # in fact freshly built
        if cached_value is None:
            self.state = c.State.MISSING
        elif cached_value == self.value:
            self.state = c.State.AS_BEFORE
        else:
            self.state = c.State.HAS_CHANGED
        # b.debug(f"{self.__class__.__name__}({self.name}).state = {self.state}")

    def do_build(self):
        # self.value is already set to the current value by check_existing_resource
        self.handle_value_and_state(self.value)


class Toc(FreshPiece):
    """HTML for the toc sidebar of a Part."""
    FRESH_ATTR = 'toc'


class LinkslistBottom(FreshPiece):  # TODO 2: integrate in the build
    """HTML for the assumedBy/requiredBy links of a Task."""
    FRESH_ATTR = 'linkslist_bottom'


class Topmatter(Piece):
    """Metadata from a Part file. Byproduct: Content."""
    CACHED_TYPE = 'dict'  # which kind of value is in the cache
    
    def __init__(self, name: str, **kwargs):
        super().__init__(name, **kwargs)
        self.add_dependency(self.directory.get_the(Sourcefile, self.sourcefile))

    def do_build(self):
        # ----- obtain file contents:
        SEPARATOR = "---\n"
        text = b.slurp(self.sourcefile)
        if SEPARATOR not in text:
            b.error(f"triple-dash separator is missing", file=self.sourcefile)
            return
        topmatter_text, content_text = text.split(SEPARATOR, 1)
        # ----- parse metadata:
        try:
            value = yaml.safe_load(topmatter_text) or dict()  # avoid None for empty topmatter
        except yaml.YAMLError as exc:
            b.error(f"metadata YAML is malformed: {str(exc)}", file=self.sourcefile)
            value = dict()  # use empty metadata as a weak replacement
        # ----- use the pieces, actual change has not necessarily happened:
        self.handle_value_and_state(value)
        content_elem = self.directory.get_the(Content, self.name)  # our byproduct
        content_elem.handle_value_and_state(content_text)


class Outputfile(Product):  # abstract class
    """Superclass for Products ending up in one file per targetdir."""
    targetdir_s: str  # student dir, if any
    targetdir_i: str  # instructor dir, if any

    @property
    def outputfile(self) -> str:  # but different for Parts
        return self.name

    @property
    def outputfile_s(self) -> str:
        return os.path.join(self.targetdir_s, self.outputfile)

    @property
    def outputfile_i(self) -> str:
        return os.path.join(self.targetdir_i, self.outputfile)

    def check_existing_resource(self):
        if not os.path.exists(self.outputfile_s) or not os.path.exists(self.outputfile_i):
            self.state = c.State.MISSING
        else:
            self.state = c.State.AS_BEFORE


class CopiedFile(Outputfile):
    """For resources which are copied verbatim. The data lives in the file system, hence no value."""
    def __init__(self, name: str, **kwargs):
        super().__init__(name, **kwargs)
        self.add_dependency(self.directory.get_the(Sourcefile, self.sourcefile))

    def do_build(self):
        b.debug(f"copying '{self.sourcefile}'\t-> '{self.targetdir_s}'")
        shutil.copy(self.sourcefile, self.outputfile_s)
        b.debug(f"copying '{self.sourcefile}'\t-> '{self.targetdir_i}'")
        shutil.copy(self.sourcefile, self.outputfile_i)


class Part(Outputfile):  # abstract class for Course, Chapter, Taskgroup, Task and their Builders, see course.py
    """Outputs identified by a name, possible targets of PARTREF."""
    TOC_LEVEL = 0  # indent level in table of contents
    title: str  # title: value
    parttype: dict[str, type['Part']]  # for using X vs XBuilder for sub-Parts
    parent: 'Part'
    course: 'Course'

    def __init__(self, name: str, **kwargs):
        super().__init__(name, **kwargs)
        self._inherit_from_parent()

    @property
    def outputfile(self) -> str:
        return f"{self.name}.html"

    @property
    def toc(self) -> str:
        return "((TOC))"  # override in concrete Part classes

    def structure_path(self) -> list['Part']:
        """List of nested parts, from a given part up to the course."""
        import sdrl.course
        structure = self
        path = []
        if isinstance(structure, sdrl.course.Task):
            path.append(structure)
            structure = structure.taskgroup
        if isinstance(structure, sdrl.course.Taskgroup):
            path.append(structure)
            structure = structure.chapter
        if isinstance(structure, sdrl.course.Chapter):
            path.append(structure)
            structure = structure.course
        if isinstance(structure, sdrl.course.Course):
            path.append(structure)
        return path

    def _inherit_from_parent(self):
        """
        Except for Course, non-Part Elements receive a parent= argument from which they copy certain
        attributes, some depending on whether this is author mode or not.
        Inheritance in the part= style of Element is not used for Parts and
        so Parts need to define sourcefile themselves.
        """
        if getattr(self, 'parent', None):  # inherit parent attrs, to make X and XBuilder simpler:
            self.course = self.parent.course
            self.parttype = self.parent.parttype
            if getattr(self.parent, 'directory', None):
                self.directory = self.parent.directory
            if getattr(self.parent, 'targetdir_s', None):
                self.targetdir_s = self.parent.targetdir_s
                self.targetdir_i = self.parent.targetdir_i


class Zipfile(Part):
    """xy.zip Outputfiles that are named Parts, plus the exceptional case itree.zip."""
    instructor_only: bool = False
    
    def __init__(self, name: str, **kwargs):
        super().__init__(name, **kwargs)
        self.add_dependency(self.directory.get_the(Zipdir, self.sourcefile))

    @property
    def innerpath(self) -> str:
        return self.name[:-len(".zip")]  # e.g. myzipdir

    @property
    def outputfile(self) -> str:  # like for non-Parts
        return self.name

    @property
    def to_be_skipped(self) -> bool:
        return False  # TODO 3: should be skipped if no [PARTREF] to it exists, unless instructor_only

    def check_existing_resource(self):
        s_ok = os.path.exists(self.outputfile_s) or self.instructor_only
        i_ok = os.path.exists(self.outputfile_i)
        self.state = c.State.AS_BEFORE if s_ok and i_ok else c.State.MISSING

    def do_build(self):
        if self.instructor_only:  # suppress printing instructor file for the normal pairs
            b.info(self.outputfile_i)
        with zipfile.ZipFile(self.outputfile_i, mode='w', 
                             compression=zipfile.ZIP_DEFLATED) as archive:  # prefer deflate for build speed
            self._zip_the_files(archive)
        if not self.instructor_only:
            b.info(self.outputfile_s)
            shutil.copy(self.outputfile_i, self.outputfile_s)
    
    def _zip_the_files(self, archive: zipfile.ZipFile):
        assert os.path.exists(self.sourcefile), f"'{self.sourcefile}' is missing!"
        for dirpath, dirnames, filenames in os.walk(self.sourcefile):
            for filename in sorted(filenames):
                sourcename = f"{dirpath}/{filename}"
                targetname = self._path_in_zip(sourcename)
                archive.write(sourcename, targetname)

    def _path_in_zip(self, sourcename: str) -> str:
        """
        Remove outside-of-zipdir prefix, then use innerpath plus inside-of-zipdir remainder.
        """
        namepos = sourcename.find(self.name)  # will always exist
        remainder = sourcename[namepos+len(self.name)+1:]
        return f"{self.innerpath}/{remainder}"


class Source(Element):  # abstract class
    """
    Abstract superclass for an input for the build.
    A Source only provides a state, not a Product, hence no build action and no value.
    The data resides either in the file system or is supplied upon instantiation. 
    """
    def do_build(self):
        pass  # Sources need no building, only checking


class Sourcefile(Source):
    """A Source that consists of a single file. Its name is the sourcefile's full path."""
    posthoc: bool  # flag "this object did not yet exist during the Sourcefile.build() phase" 

    def __init__(self, name, **kwargs):
        super().__init__(name, **kwargs)
        self.sourcefile = self.name
        if getattr(self, 'posthoc', False):
            # whenever a Sourcefile is created when it is found as an includefile in a markdown render 
            # (i.e. on first build or for a new includefile), posthoc=True is passed to the constructor.
            # In this case, we must do cache.record_file() or else the file would be 
            # considered new (and may trigger lots of rebuild actions) during the next build.
            # For direct Sourcefiles of Parts, however, we must not do this, so that 
            # the fact that the file was never seen before can be used to set state HAS_CHANGED 
            # in check_existing_resource().
            self.cache.record_file(self.sourcefile, self.cache_key)

    def check_existing_resource(self):
        self.state = self.cache.filestate(self.sourcefile, self.cache_key)
        if self.state == c.State.MISSING:  # we have never encountered this file in our cache history
            self.cache.record_file(self.sourcefile, self.cache_key)
            self.state = c.State.HAS_CHANGED


class Zipdir(Source):
    """A directory tree *.zip/** to be turned into a same-named ZIP file."""

    def __init__(self, zipdirpath: str, **kwargs):
        super().__init__(zipdirpath, **kwargs)
        assert zipdirpath[-1] != '/'  # dirprefix must not end with a slash, else our logic would break
        self.sourcefile = zipdirpath  # e.g. ch/mychapter/mytaskgroup/myzipdir.zip 
        self.title = os.path.basename(zipdirpath)  # e.g. myzipdir.zip

    def check_existing_resource(self):
        """HAS_CHANGED if any file in the tree is new or the set of filenames differs from cache."""
        # ----- analyze Zipdir tree:
        current_filelist = []
        new_file_found = False
        for dirpath, dirnames, filenames in os.walk(self.sourcefile):
            for filename in sorted(filenames):
                sourcename = f"{dirpath}/{filename}"
                if self.cache.is_recent(sourcename):
                    new_file_found = True
                current_filelist.append(sourcename)
        current_fileset = set(current_filelist)
        # ----- determine state according to changes and cache:
        old_fileslist, cache_state = self.cache.cached_list(self.cache_key)
        old_fileset = set(old_fileslist or [])
        filesets_differ = old_fileset != current_fileset
        if cache_state == c.State.MISSING or new_file_found or filesets_differ:
            b.debug(f"Zipdir.check({self.name}): cache {cache_state}, new_file {new_file_found}, "
                    f"filesets differ {filesets_differ}")
            self.cache.write_list(self.cache_key, current_filelist)
            self.state = c.State.HAS_CHANGED
        else:
            self.state = c.State.AS_BEFORE


class Step(Element):
    """Pseudo-element that does not represent data, but computation."""
    def check_existing_resource(self):
        self.state = c.State.HAS_CHANGED  # do_build() must always be called.



"""Combined Elements registry/factory and build orchestrator."""
import itertools

import base as b
import typing as tg


class Directory:
    """
    Each Element is registered here and can be accessed by type and name. 
    The Builder knows when to build each and builds them in an order such that 
    when a Product gets built, the state of all its dependencies is already known.
    This order is also fully determined by the Element types alone.
    """
    def __init__(self, cache):
        import sdrl.elements as el
        import sdrl.course as course
        import sdrl.glossary as glossary
        self.cache = cache
        self.managed_types = [
            # Each has a downcased dict attribute use by get_the()/make_the().
            # The ordering is the build ordering:
            el.Sourcefile, el.CopiedFile,
            el.Zipdir, el.Zipfile,
            el.Topmatter, el.Content, course.MetadataDerivation,
            el.IncludeList_s, el.IncludeList_i, el.TermrefList,
            el.Body_s, el.Body_i, el.Glossarybody,
            el.Toc, el.LinkslistBottom,
            course.Course, course.Chapter, course.Taskgroup, course.Task, glossary.Glossary,
        ]
        for thistype in self.managed_types:
            dictname = thistype.__name__.lower()
            setattr(self, dictname, dict())

    def get_the(self, mytype: type, name: str) -> 'sdrl.elements.Element':
        """Retrieve existing object from the directory."""
        the_dict = self._getdict(mytype)
        return the_dict.get(name)

    def make_the(self, mytype: type, name: str, **kwargs) -> 'sdrl.elements.Element':
        """Instantiate object and store it in the directory. Must be a new entry."""
        the_dict = self._getdict(mytype)
        if name in the_dict:
            b.debug(f"make_the: overwriting internal entry {mytype.__name__}({name}) from {b.caller(1)}")
        instance = mytype(name, directory=self, **kwargs)
        the_dict[name] = instance
        return instance

    def take_the(self, mytype: type, name: str, instance):
        """Store the existing element in the directory. Must be a new entry."""
        the_dict = self._getdict(mytype)
        if name in the_dict:
            b.debug(f"take_the: overwriting internal entry {mytype.__name__}({name})")
        the_dict[name] = instance

    def make_or_get_the(self, mytype: type, name: str, **kwargs) -> 'sdrl.elements.Element':
        instance = self.get_the(mytype, name)
        return instance if instance else self.make_the(mytype, name, **kwargs)

    def record_the(self, mytype: type, name, instance):
        """Store existing object into the directory."""
        the_dict = self._getdict(mytype)
        if name in the_dict and the_dict[name] is not instance:
            b.debug(f"record_the: overwriting internal entry {mytype.__name__}({name})")
        the_dict[name] = instance

    def build(self):
        alldicts = ((mytype, self._getdict(mytype)) for mytype in self.managed_types)
        for thistype, thisdict in alldicts:
            b.debug(f"building all Elements of type {thistype.__name__}")
            for elem in thisdict.values():
                elem.build()

    def get_all(self, what: type | str) -> tg.Iterable:
        """All entries with a given type or with a given name (in any type)."""
        if isinstance(what, type):
            return self._getdict(what).values()
        # collect all elements with name what:
        result = []
        for typ in self.managed_types:
            candidate = self._getdict(typ).get(what, None)
            if candidate:  # skip types with no such entry
                result.append(candidate)
        return result

    def get_all_outputfiles(self) -> tg.Iterator:
        import sdrl.elements as el
        iterators = [self.get_all(t) for t in self.managed_types
                     if issubclass(t, el.Outputfile)]
        return itertools.chain(*iterators)

    def _getdict(self, thetype: type):
        dictname = thetype.__name__.lower()
        return getattr(self, dictname)



"""Element cache for incremental build."""
import dbm
import enum
import itertools
import json
import os
import time
import typing as tg
import zlib

import base as b


UNCOMPRESSED_LIMIT = 40  # length of longest string to store uncompressed
ZLIB_WBITS = -15
Cacheable = str | list[str] | b.StrAnyDict  # what can be put in the cache
CacheEntryType = None | Cacheable  # what cache queries can return


class State(enum.StrEnum):
    """
    State of an Element wrt a Product or as a cache entry.
    missing is a non-file/non-cacheentry: must build.
    has_changed is a younger file or freshly written cache entry: must build or have built.
    as_before is an old file or not-overwritten cache entry or outputfile: need not build or have not built.
    """
    MISSING = 'MISSING'  # for Source or before build: must build; impossible after build
    HAS_CHANGED = 'HAS_CHANGED'  # for Source or before build: must build; after build: have built
    AS_BEFORE = 'AS_BEFORE'  # for Source or before build: need not build; after build: have not built


class SedrilaCache:
    """
    All intermediate products and some config values are stored in the cache.
    There are four entry types:
    - files are reflected as a cache key with empty value.
      The has_changed reference time for files is stored in a single entry TIMESTAMP_KEY.
    - str are just that.
    - list[str] and set[str] are stored as a string using LIST_SEPARATOR.
    - dict[Any] are stored as json.
    Keys take the form partname__entrytype.
    Helper entries use __dunder__ keys.
    The actual filename(s) depend(s) on the chosen DBM implementation, which depends on
    what is installed on the system, which means cache files are not necessarily portable
    to other computers.  The default plain-Python DBM will have a single *.db file.
    """
    LIST_SEPARATOR = '|'  # separates entries in list-valued dbm entries. Symbol is forbidden in all names.
    TIMESTAMP_KEY = '__mtime__'  # unix timestamp: seconds since epoch
    DIRTYFILES_KEY = '__dirtyfileslist__'  # previous_dirtyfiles

    db: dict  # in fact a dbm._Database
    persistent_mode: bool  # non-persistent mode for testing/student/instructor via cache_filename=""
    written: b.StrAnyDict  # what was written into cache since start
    timestamp_start: int  # when did the current build process begin -> the future reference time
    timestamp_cached: int  # when did the previous build process begin -> the current reference time
    previous_dirtyfiles: set[str]  # files marked dirty during last run
    new_dirtyfiles: set[str]  # files marked dirty during present run

    def __init__(self, cache_filename: str, start_clean: bool):
        # self.timestamp_start = int(time.time() + 0.5)  # add half second in case filesystem mtime rounds, not truncates
        self.timestamp_start = round(time.time(), 3)  # milliseconds suffice for us
        self.persistent_mode = bool(cache_filename)
        if self.persistent_mode:
            self.db = dbm.open(cache_filename, flag='n' if start_clean else 'c')  # open or create dbm file
        else:
            self.db = dict()
        self.written = dict()
        tsk = self.TIMESTAMP_KEY
        timestamp_cached = float(_decompress(self.db[tsk]) if tsk in self.db else "0")  # default: everything is old
        self.timestamp_cached = round(timestamp_cached, 3)  # milliseconds suffice for us
        b.debug(f"cache.mtime/timestamp_cached: {self.timestamp_cached}")
        dirtyfiles, dirtyfiles_state = self.cached_list(self.DIRTYFILES_KEY)
        self.previous_dirtyfiles = set(dirtyfiles)
        self.new_dirtyfiles = set()
        # self._dump(limit=256)  # debug, if needed

    def __contains__(self, key: str) -> bool:
        return key in self.written or key in self.db

    def __getitem__(self, key: str) -> tg.Any:
        if key in self.written:
            return self.written[key]
        elif key in self.db:
            return _decompress(self.db[key])
        raise ValueError(key)

    def __setitem__(self, key: str, value: tg.Any):
        self.written[key] = value

    def __getstate__(self):  # for pickle
        return None  # SedrilaCache has no state that should be pickled

    def __setstate__(self, state):  # for unpickle
        pass  # no useful state is restored upon unpickle; one must create a new instance
    
    @property
    def mtime(self) -> int:
        return self.timestamp_cached

    def cached_str(self, key: str) -> tuple[str, State]:
        return self._entry(key, self._as_is)

    def cached_list(self, key: str) -> tuple[list[str], State]:
        return self._entry(key, self._as_list)

    def cached_set(self, key: str) -> tuple[set[str], State]:
        return self._entry(key, self._as_set)

    def cached_dict(self, key: str) -> tuple[b.StrAnyDict, State]:
        return self._entry(key, self._as_dict)

    def filestate(self, pathname: str, cache_key: str) -> State:
        """
        Non-existing file: MISSING.
        Before-unseen file: MISSING. **Special case!`**
        Dirty file, or recent file: HAS_CHANGED.
        Otherwise (file with old time): AS_BEFORE.
        """
        if not os.path.exists(pathname):
            return State.MISSING
        cache_state = self.state(cache_key)
        if cache_state == State.MISSING:
            # b.debug(f"{pathname} not in cache")
            return State.MISSING
        if self.is_recent(pathname) or self.is_dirty(pathname):
            # b.debug(f"{pathname} has younger mtime")
            return State.HAS_CHANGED
        else:
            return State.AS_BEFORE

    def record_file(self, path: str, cache_key: str):
        assert cache_key not in self.written  # we should usually write everything only once
        self.written[cache_key] = ""  # file entries are empty because the file itself holds the data

    def is_recent(self, pathname: str) -> bool:
        """Whether pathname's mtime is larger than the cache's global mtime."""
        filetime = os.stat(pathname).st_mtime
        it_is = filetime > self.mtime
        if it_is:
            b.debug(f"cache.is_recent({pathname}): {filetime} > {self.mtime}")
        return it_is

    def is_dirty(self, pathname: str) -> bool:
        """Whether pathname was marked dirty in previous run."""
        it_is = pathname in self.previous_dirtyfiles
        if it_is:
            b.debug(f"cache.is_dirty({pathname})!")
        return it_is

    def set_file_dirty(self, filename: str):
        """
        When a file produces an error or warning message, we mark it as 'dirty'.
        On the next run, the cache will flag such files as HAS_CHANGED, so that they are built
        again and the message is produced again. 
        """
        b.debug(f"cache.set_file_dirty({filename})")
        self.new_dirtyfiles.add(filename)

    def write_str(self, key: str, value: str):
        assert key not in self.written  # we should write everything only once
        self.written[key] = value

    def write_list(self, key: str, value: list[str]):
        assert key not in self.written  # we should write everything only once
        self.written[key] = value

    def write_set(self, key: str, value: set[str]):
        assert key not in self.written  # we should write everything only once
        self.written[key] = value

    def write_dict(self, key: str, value: b.StrAnyDict):
        assert key not in self.written  # we should write everything only once
        self.written[key] = value

    def close(self):
        """Bring the persistent cache file up-to-date and close dbm."""
        converters = {str: self._as_is, list: self._from_list, set: self._from_set, dict: self._from_dict}
        self.db[self.TIMESTAMP_KEY] = _compress(str(self.timestamp_start))  # update mtime
        self.write_list(self.DIRTYFILES_KEY, list(self.new_dirtyfiles))
        for key, value in self.written.items():
            if value is None:  # should not happen
                b.debug(f"cache['{key}'] is None")
                continue
            converter = converters[type(value)]
            data = converter(value)
            self.db[key] = _compress(data)
        if self.persistent_mode:
            self.db.__exit__()  # dbm file context manager operation 

    def state(self, key: str) -> State:
        if key in self.written:
            return State.HAS_CHANGED
        elif key in self.db:
            return State.AS_BEFORE
        else:
            return State.MISSING

    @staticmethod
    def _as_is(e: str) -> str:
        return e

    def _as_list(self, e: str) -> list[str]:
        return e.split(self.LIST_SEPARATOR) if e else []

    def _as_set(self, e: str) -> set[str]:
        return set(e.split(self.LIST_SEPARATOR)) if e else set()

    @staticmethod
    def _as_dict(e: str) -> b.StrAnyDict:
        return json.loads(e) if e else dict()

    def _from_list(self, e: list[str]) -> str:
        return self.LIST_SEPARATOR.join(e)

    def _from_set(self, e: set[str]) -> str:
        return self.LIST_SEPARATOR.join(e)

    @staticmethod
    def _from_dict(e: b.StrAnyDict) -> str:
        return json.dumps(e, indent=2, check_circular=False)

    def _entry(self, key: str, converter: tg.Callable[[str], CacheEntryType]) -> tuple[CacheEntryType, State]:
        """The only internal cache accessor function"""
        if key in self.written:
            return (self.written[key], State.HAS_CHANGED)
        elif key in self.db:
            return (converter(_decompress(self.db[key])), State.AS_BEFORE)
        else:
            return (converter(None), State.MISSING)

    def _dump(self, limit: int):
        keys = sorted(self.db.keys())
        for key in keys:
            value = _decompress(self.db[key])
            print(f"{key}:\t{value[:limit]}")

    def _storefilelist(self, key: str, samefiles: list[str], newfiles: list[str]):
        thelist = self.LIST_SEPARATOR.join(itertools.chain(samefiles, newfiles))
        self.db[key] = thelist


def _compress(s: str) -> bytes:
    """Encoded, gzip-compressed string; short strings stay uncompressed with a 0 indicator byte prepended."""
    if len(s) > UNCOMPRESSED_LIMIT:
        return zlib.compress(s.encode(), level=6, wbits=ZLIB_WBITS)
    else:
        return b"\x00" + s.encode()


def _decompress(b: bytes) -> str:
    if b[0] == 0:
        return b[1:].decode()
    else:
        return zlib.decompress(b, wbits=ZLIB_WBITS, bufsize=64000).decode()



"""sedrila-specific HTML generation helper routines."""
import sdrl.constants as c


difficulty_levels = ('verylow', 'low', 'medium', 'high')

def as_attribute(text: str) -> str:
    """Cleans text so that it can appear between double quotes in an HTML attribute."""
    return text.replace('"', "'").replace("\n", " ")  # no doublequotes, no line breaks


def breadcrumb(*args):
    """Renders breadcrumb HTML fragment from list of items with breadcrumb_item property."""
    SEPARATOR = " &gt; "
    return '<nav><span class="breadcrumbs">%s</span></nav>' % SEPARATOR.join([arg.breadcrumb_item for arg in args])


def difficulty_symbol(level: int) -> str:
    difficulty_text = difficulty_levels[level - 1]
    diffclass = f"class='difficulty{level}'"
    circle = f"<span {diffclass} title='Difficulty: {difficulty_text}'>{c.HTML_DIFFICULTY_SIGN}</span>"
    return circle


def indented_block(text: str, level: int, classes: str) -> str:
    return "".join([(level+1) * "  ", f"<div class='indent{level} {classes}'>", text, "</div>"])




"""
Represent and handle the contents of SeDriLa: Course, Chapter, Taskgroup, Task.
There are two ways how these objects can be instantiated:
In 'author' mode, metadata comes from sedrila.yaml and the partfiles.
Otherwise, metadata comes from METADATA_FILE. 
"""
import dataclasses
import functools
import glob
import graphlib
import itertools
import numbers
import os
import re
import typing as tg

import base as b
import sdrl.constants as c
import sdrl.elements as el
import sdrl.glossary as glossary
import sdrl.html as h
import sdrl.macros as macros
import sdrl.partbuilder

sedrila_libdir = os.path.dirname(os.path.dirname(__file__))  # either 'py' (for dev install) or top-level
if sedrila_libdir.endswith('py'):  # we are a dev install and must go one more level up:
    sedrila_libdir = os.path.dirname(sedrila_libdir)


class Task(el.Part):
    DIFFICULTY_RANGE = range(1, len(h.difficulty_levels) + 1)
    TOC_LEVEL = 2  # indent level in table of contents
    
    course: 'Course'
    timevalue: tg.Union[int, float]  # task timevalue: (in hours)
    difficulty: int  # difficulty: int from DIFFICULTY_RANGE
    assumes: list[str] = []  # tasknames: This knowledge is assumed to be present
    requires: list[str] = []  # tasknames: These specific results will be reused here
    workhours: float = 0.0  # time student has worked on this according to commit msgs
    is_accepted: bool = False  # whether instructor has ever marked it 'accept'
    rejections: int = 0  # how often instructor has marked it 'reject'

    taskgroup: 'Taskgroup'  # where the task belongs

    def __init__(self, name: str, **kwargs):
        super().__init__(name, **kwargs)
        self.course.namespace_add(self)

    def __hash__(self) -> int:
        return hash(self.name)

    @property
    def allowed_attempts(self) -> int:
        course = self.taskgroup.chapter.course
        return int(course.allowed_attempts_base + course.allowed_attempts_hourly*self.timevalue)

    @property
    def remaining_attempts(self) -> int:
        return max(0, self.allowed_attempts - self.rejections)

    @property
    def sourcefile(self) -> str:
        return f"{self.course.chapterdir}/{self.taskgroup.chapter.name}/{self.taskgroup.name}/{self.name}.md"

    def from_json(self, task: b.StrAnyDict) -> 'Task':
        """Initializer used in student and instructor mode. Ignores all additional attributes."""
        self.taskgroup = tg.cast('Taskgroup', self.parent)
        self.title = task['title']
        self.timevalue = task['timevalue']
        self.difficulty = task['difficulty']
        self.assumes = task['assumes']
        self.requires = task['requires']
        return self


@functools.total_ordering
class Taskbuilder(sdrl.partbuilder.PartbuilderMixin, Task):
    TEMPLATENAME = "task.html"
    explains: list[str] = []  # terms (for backlinks in glossary)
    assumed_by: list[str] = []  # tasknames: inverse of assumes
    required_by: list[str] = []  # tasknames: inverse of requires

    taskgroup: 'Taskgroupbuilder'  # where the task belongs

    def __init__(self, name: str, **kwargs):
        super().__init__(name, **kwargs)
        self.directory.record_the(Task, self.name, self)
        self.make_std_dependencies(use_toc_of=self.taskgroup)
        self.make_dependency(el.LinkslistBottom, part=self)

    def __eq__(self, other):
        return other.name == self.name

    def __lt__(self, other):
        if not isinstance(other, type(self)):
            return NotImplemented
        return (self.difficulty < other.difficulty or
                (self.difficulty == other.difficulty and self.name < other.name))

    def __hash__(self) -> int:
        return hash(self.name)

    @property
    def allowed_attempts(self) -> int:
        course = self.taskgroup.chapter.course
        return int(course.allowed_attempts_base + course.allowed_attempts_hourly*self.timevalue)

    @functools.cached_property
    def linkslist_bottom(self) -> str:
        return self._render_task_linkslist('assumed_by', 'required_by')

    @functools.cached_property
    def linkslist_top(self) -> str:
        return self._render_task_linkslist('assumes', 'requires')

    @property
    def remaining_attempts(self) -> int:
        return max(0, self.allowed_attempts - self.rejections)

    @property
    def to_be_skipped(self) -> bool:
        return (self.skipthis or
                self.taskgroup.to_be_skipped or self.taskgroup.chapter.to_be_skipped)

    @property
    def toc_link_text(self) -> str:
        href = f"href='{self.outputfile}'"
        titleattr = f"title=\"{self.title}\""
        diffsymbol = h.difficulty_symbol(self.difficulty)
        timevalue = ("<span class='timevalue-decoration' title='Timevalue: %s hours'>%s</span>" %
                     (self.timevalue, self.timevalue))
        refs = (self._taskrefs('assumed_by') + self._taskrefs('required_by') +
                self._taskrefs('assumes') + self._taskrefs('requires'))
        return f"<a {href} {titleattr}>{self.name}</a> {diffsymbol} {timevalue}{refs}"

    @property
    def toc(self) -> str:
        return self.taskgroup.toc

    def as_json(self) -> b.StrAnyDict:
        return dict(name=self.name, slug=self.name,  # slug for backwards compatibility, TODO 3: remove 2025-01
                    title=self.title, timevalue=self.timevalue, difficulty=self.difficulty,
                    assumes=self.assumes, requires=self.requires)

    def process_topmatter(self, sourcefile: str, topmatter: b.StrAnyDict, course: 'Coursebuilder'):
        b.copyattrs(sourcefile,
                    topmatter, self,
                    mustcopy_attrs='title, timevalue, difficulty',
                    cancopy_attrs='stage, explains, assumes, requires',
                    mustexist_attrs='',
                    typecheck=dict(timevalue=numbers.Number, difficulty=int))
        self.evaluate_stage(sourcefile, course)

        # ----- ensure explains/assumes/requires are lists:
        def _handle_strlist(attrname: str):
            attrvalue = getattr(self, attrname, None)
            if isinstance(attrvalue, str):
                setattr(self, attrname, re.split(r", *", attrvalue))
            elif not attrvalue:
                setattr(self, attrname, [])
            else:
                b.error(f"value of '{attrname}:' must be a (non-empty) string", file=sourcefile)
                setattr(self, attrname, [])

        _handle_strlist('explains')
        _handle_strlist('assumes')
        _handle_strlist('requires')
        # ----- add to glossary:
        if self.explains:
            for term in self.explains:
                course.glossary.explains(self.name, term)
        # ----- done:
        return self

    @classmethod
    def expand_diff(cls, call: macros.Macrocall) -> str:  # noqa
        assert call.macroname == "DIFF"
        level = int(call.arg1)
        diffrange = cls.DIFFICULTY_RANGE
        if level not in diffrange:
            call.error(f"Difficulty must be in range {min(diffrange)}..{max(diffrange)}")
            return ""
        return h.difficulty_symbol(level)

    def _taskrefs(self, attr_name: str) -> str:
        """Create a toc link dedoration for one set of related tasks."""
        attr_cssclass = "%s-decoration" % attr_name.replace("_", "-")
        attr_label = attr_name.replace("_", " ")
        refslist = getattr(self, attr_name)
        if len(refslist) == 0:
            return ""
        title = "%s: %s" % (attr_label, ", ".join(refslist))
        return ("<span class='%s' title='%s'>%s</span>" %
                (attr_cssclass, title, ""))  # label is provided by CSS

    def _render_task_linkslist(self, a_attr: str, r_attr: str) -> str:
        """HTML for the links to assumes/requires (or assumed_by/required_by) related tasks on a task page."""
        links = []
        a_links = sorted((f"[PARTREF::{part}]" for part in getattr(self, a_attr)))
        r_links = sorted((f"[PARTREF::{part}]" for part in getattr(self, r_attr)))
        a_cssname = a_attr.replace("_", "")
        r_cssname = r_attr.replace("_", "")
        any_links = a_links or r_links
        if any_links:
            links.append(f"\n<aside class='{a_cssname}-{r_cssname}-linkblock'>\n")
        if a_links:
            links.append(f" <div class='{a_cssname}-links'>\n   ")
            links.append("  " + macros.expand_macros("-", self.name, ", ".join(a_links)))
            links.append("\n </div>\n")
        if r_links:
            links.append(f" <div class='{r_cssname}-links'>\n")
            links.append("  " + macros.expand_macros("-", self.name, ", ".join(r_links)))
            links.append("\n </div>\n")
        if any_links:
            links.append("</aside>\n")
        return "".join(links)


class Course(el.Part):
    """
    Abstract superclass for the master data object for this run. Can be instantiated in two different forms: 
    - as Coursebuilder (for author): 
      From a handwritten YAML file, called configfile. 
      Will then read all content for a build and employ directory and cache.
    - as CourseSI (for student and instructor): 
      From a JSON metadata file generated during build.
    The present superclass contains the attributes and logic that both have in common.
    """
    MUSTCOPY_ADDITIONAL = '???'  # set in subclasses
    CANCOPY_ADDITIONAL = '???'  # set in subclasses

    context: str  # for error messages: where config data comes from
    configdict: b.StrAnyDict  # all data from c.CONFIGFILE or c.METADATAFILE
    namespace: dict[str, el.Part]  # partname -> part, for obtaining Parts in any of the logic

    instructors: list[b.StrAnyDict]
    init_data: b.StrAnyDict = {}
    allowed_attempts: str  # "2 + 0.5/h" or so, int+decimal, h is the task timevalue multiplier
    allowed_attempts_base: int  # the int part of allowed_attempts
    allowed_attempts_hourly: float  # the decimal part of allowed_attempts

    def __init__(self, **kwargs):
        super().__init__("...", **kwargs)  # preliminary name!
        self.course = self
        self.namespace = dict()
        self._read_config(self.configdict)
        self.allowed_attempts_base, self.allowed_attempts_hourly = self._parse_allowed_attempts()


    @functools.cached_property  # beware: call this only once initialization is complete!
    def taskdict(self) -> dict[str, Task|Taskbuilder]:
        return {k:v for k,v in self.namespace.items() if isinstance(v, Task)}

    def get_part(self, context: str, partname: str) -> el.Part:
        """Return part for given partname or self (and print error msg) if no such part exists."""
        if partname in self.namespace:
            return self.namespace[partname]
        b.error(f"part '{partname}' does not exist", file=context)
        return self

    def namespace_add(self, newpart: el.Part):
        name = newpart.name
        if name in self.namespace:
            b.critical("Files '%s' and '%s':\n   part name collision" %  # critical because it makes assumptions wrong
                       (self._partpath(self.namespace[name]), self._partpath(newpart)))
        else:
            self.namespace[name] = newpart

    def task(self, taskname: str):
        """Return Task for given taskname or None if no such task exists."""
        return self.taskdict.get(taskname, None)  # noqa

    def _parse_allowed_attempts(self) -> tuple[int, float]:
        mm = re.match(r"(?P<base>\d+)(\s?\+\s?(?P<time>\d+\.\d+)\s?/\s?h)?", self.allowed_attempts)
        if not mm:
            msg1 = f"'allowed_attempts' must have a format exactly like '2 + 0.5/h' or '2'"
            b.error(f"{msg1}, not '{self.allowed_attempts}'", file=getattr(self, 'sourcefile', ''))
            return 2, 0
        return int(mm.group("base")), float(mm.group("time") or "0")

    @staticmethod
    def _partpath(p: el.Part) -> str:
        fullpath = p.sourcefile
        basename = os.path.basename(fullpath)
        dirname = os.path.dirname(fullpath)
        if basename == "index.md":
            return dirname
        else:
            return fullpath

    def _read_config(self, configdict: b.StrAnyDict):
        b.copyattrs(self.context,
                    configdict, self,
                    mustcopy_attrs=('title, name, instructors, allowed_attempts' +
                                    self.MUSTCOPY_ADDITIONAL),
                    cancopy_attrs=('init_data' +
                                   self.CANCOPY_ADDITIONAL),
                    mustexist_attrs='chapters',
                    report_extra=bool(self.MUSTCOPY_ADDITIONAL))


class CourseSI(Course):
    MUSTCOPY_ADDITIONAL = ''
    CANCOPY_ADDITIONAL = ''

    chapters: list['Chapter']

    def __init__(self, configdict: b.StrAnyDict, context: str):
        super().__init__(configdict=configdict, context=context)
        self.parttype = dict(Chapter=Chapter, Taskgroup=Taskgroup, Task=Task)
        self._init_parts(self.configdict)
        
    def _init_parts(self, configdict: dict):
        self.chapters = [Chapter(ch['name'], parent=self, chapterdict=ch)  # noqa
                         for ch in configdict['chapters']]



class Coursebuilder(sdrl.partbuilder.PartbuilderMixin, Course):
    """Course with the additions required for author mode. (Chapter, Taskgroup, Task have both in one.)"""
    MUSTCOPY_ADDITIONAL = ', chapterdir, altdir, stages'
    CANCOPY_ADDITIONAL = ', baseresourcedir, itreedir, templatedir, blockmacro_topmatter, htaccess_template'
    TEMPLATENAME = "homepage.html"

    include_stage: str  # lowest stage that parts must have to be included in output
    targetdir_s: str  # where to render student output files
    targetdir_i: str  # where to render instructor output files

    baseresourcedir: str = f"{sedrila_libdir}/baseresources"
    chapterdir: str
    altdir: str
    itreedir: str | None
    templatedir: str = f"{sedrila_libdir}/templates"
    blockmacro_topmatter: dict[str, str]
    htaccess_template: str = None  # structure of .htaccess file generated in instructor website
    stages: list[str]  # list of allowed values of stage in parts 

    course: 'Coursebuilder'
    chapters: list['Chapterbuilder']
    include_stage_index: int  # index in stages list, or len(stages) if include_stage is ""
    mtime: float  # in READ cache mode: tasks have changed if they are younger than this
    taskorder: list[Taskbuilder]  # If task B assumes or requires A, A will be before B in this list.
    glossary: glossary.Glossary

    def __init__(self, *, configfile: str, **kwargs):
        self.configfile = self.context = configfile
        self.configdict = b.slurp_yaml(configfile)
        super().__init__(**kwargs)
        self.parttype = dict(Chapter=Chapterbuilder, Taskgroup=Taskgroupbuilder, Task=Taskbuilder)
        self._read_config(self.configdict)
        self._init_parts(self.configdict, self.include_stage)

    @property
    def breadcrumb_item(self) -> str:
        titleattr = f"title=\"{h.as_attribute(self.title)}\""
        return f"<a href='index.html' {titleattr}>{self.name}</a>"

    @property
    def outputfile(self) -> str:
        return "index.html"

    @property
    def sourcefile(self) -> str:
        return f"{self.chapterdir}/index.md"

    @functools.cached_property
    def toc(self) -> str:
        return sdrl.partbuilder.toc(self)

    def add_inverse_links(self):
        """add Task.required_by/Task.assumed_by lists."""
        for taskname, task in self.taskdict.items():
            task.assumed_by = []  # so we do not append to the class attribute
            task.required_by = []
        for taskname, task in self.taskdict.items():
            for assumed_taskname in task.assumes:
                assumed_task = self.task(assumed_taskname)
                if assumed_task:
                    assumed_task.assumed_by.append(taskname)
            for required_taskname in task.requires:
                required_task = self.task(required_taskname)
                if required_task:
                    required_task.required_by.append(taskname)

    def as_json(self) -> b.StrAnyDict:
        result = dict(title=self.title,
                      name=self.name, breadcrumb_title=self.name,
                      instructors=self.instructors,
                      init_data=getattr(self, 'init_data', dict()),
                      allowed_attempts=self.allowed_attempts,
                      chapters=[chapter.as_json() for chapter in self.chapters])
        return result

    def check_links(self):
        for task in self.taskdict.values():
            for assumed in task.assumes:
                if not self.namespace.get(assumed, None):
                    b.error(f"assumed part '{assumed}' does not exist", file=task.sourcefile)
            for required in task.requires:
                if not self.namespace.get(required, None):
                    b.error(f"required part '{required}' does not exist", file=task.sourcefile)

    def compute_taskorder(self):
        """
        Set self.taskorder such that it respects the 'assumes' and 'requires'
        dependencies globally (across all taskgroups and chapters).
        The attribute will be used for ordering the tasks when rendering a taskgroup.
        """
        # ----- prepare dependency graph for topological sorting:
        graph = dict()  # maps a task to a set of tasks it depends on.
        for mytaskname, mytask in self.taskdict.items():
            dependencies = set()  # collection of all tasks required or assumed by mytask
            for assumedtask in mytask.assumes:
                if assumedtask in self.taskdict:
                    dependencies.add(self.taskdict[assumedtask])
            for requiredtask in mytask.requires:
                if requiredtask in self.taskdict:
                    dependencies.add(self.taskdict[requiredtask])
            graph[mytask] = dependencies
        # ----- compute taskorder (or report dependency cycle if one is found):
        self.taskorder = self._taskordering_for_toc(graph)

    @dataclasses.dataclass
    class Volumereport:
        rows: tg.Sequence[tg.Tuple[str, int, float]]
        columnheads: tg.Sequence[str]

    def volume_report_per_chapter(self) -> Volumereport:
        return self._volume_report(self.chapters, "Chapter",
                                   lambda t, c: t.taskgroup.chapter == c, lambda c: c.name)

    def volume_report_per_difficulty(self) -> Volumereport:
        return self._volume_report(Task.DIFFICULTY_RANGE, "Difficulty",
                                   lambda t, d: t.difficulty == d, lambda d: h.difficulty_levels[d-1])

    def volume_report_per_stage(self) -> Volumereport:
        return self._volume_report(self.stages + [None], "Stage",
                                   lambda t, s: t.stage == s, lambda s: s or "done", include_all=True)

    def _add_baseresources(self):
        for direntry in os.scandir(self.baseresourcedir):
            if direntry.is_file():
                self.directory.make_the(el.Sourcefile, direntry.path)
                self.directory.make_the(el.CopiedFile, direntry.name, sourcefile=direntry.path,
                                        targetdir_s=self.targetdir_s, targetdir_i=self.targetdir_i)
            else:
                b.warning("is not a plain file. Ignored.", file=direntry)

    def _collect_zipdirs(self):
        for zf in self.directory.get_all(el.Zipfile):
            self.namespace_add(zf)

    def _init_parts(self, configdict: dict, include_stage: str):
        self.directory.record_the(Course, self.name, self)
        self.namespace_add(self)
        self.make_std_dependencies(use_toc_of=self)
        # ----- handle include_stage:
        if include_stage in self.stages:
            self.include_stage = include_stage
            self.include_stage_index = self.stages.index(include_stage)
        else:
            if include_stage != '':  # empty is allowed
                b.error(f"'--include_stage {include_stage}' not allowed, must be one of {self.stages}")
            self.include_stage = ''  # include only parts with no stage
            self.include_stage_index = len(self.stages)
        # ----- create Chapters, Taskgroups, Tasks:
        self.chapters = [Chapterbuilder(ch['name'], parent=self, chapterdict=ch) 
                         for ch in configdict['chapters']]
        # ----- create Zipdirs, Glossary:
        self.find_zipdirs()
        self._collect_zipdirs()  # TODO 3: collect only what gets referenced
        self.glossary = glossary.Glossary(c.AUTHOR_GLOSSARY_BASENAME, parent=self)
        self.directory.record_the(glossary.Glossary, self.glossary.name, self.glossary)
        self.namespace_add(self.glossary)
        # ----- create MetadataDerivation and baseresources:
        self.directory.make_the(MetadataDerivation, self.name, part=self, course=self)
        self._add_baseresources()

    @staticmethod
    def _taskordering_for_toc(graph) -> list[Taskbuilder]:
        sorter = graphlib.TopologicalSorter(graph)
        sorter.prepare()
        result = []
        try:
            while sorter.is_active():
                node_group = sorted(sorter.get_ready())
                result.extend(node_group)
                sorter.done(*node_group)        
        except graphlib.CycleError as exc:
            msg = "Some tasks' 'assumes' or 'requires' dependencies form a cycle:\n"
            b.critical(msg + exc.args[1])
        return result

    def _volume_report(self, rowitems: tg.Iterable, column1head: str,
                       select: tg.Callable[[Task, tg.Any], bool],
                       render: tg.Callable[[tg.Any], str],
                       include_all=False) -> Volumereport:
        """Tuples of (category, num_tasks, timevalue_sum)."""
        result = []
        for row in rowitems:
            num_tasks = sum((1 for t in self.taskdict.values() if select(t, row) and not t.to_be_skipped))
            timevalue_sum = sum((t.timevalue for t in self.taskdict.values() if select(t, row) and not t.to_be_skipped))
            if num_tasks > 0 or include_all:
                result.append((render(row), num_tasks, timevalue_sum))
        return self.Volumereport(result, (column1head, "#Tasks", "Timevalue"))


class Chapter(el.Part):
    course: Course
    chapterdict: b.StrAnyDict
    taskgroups: list['Taskgroup']

    def __init__(self, name: str, **kwargs):
        super().__init__(name, **kwargs)
        self.course.namespace_add(self)
        context = f"chapter in {self.course.context}"
        self._init_from_dict(context, self.chapterdict)
        self._init_from_file(context, self.course)
        self.taskgroups = [self.parttype['Taskgroup'](taskgroup['name'], 
                                                      parent=self, chapter=self, 
                                                      taskgroupdict=taskgroup)
                           for taskgroup in (self.chapterdict.get('taskgroups') or [])]

    def _init_from_dict(self, context: str, chapter: b.StrAnyDict):
        b.copyattrs(context,
                    chapter, self,
                    mustcopy_attrs='name',
                    mustexist_attrs='taskgroups',
                    cancopy_attrs='title,slug')

    def _init_from_file(self, context: str, course: Coursebuilder):
        pass  # only present in builder class


class Chapterbuilder(sdrl.partbuilder.PartbuilderMixin, Chapter):
    TEMPLATENAME = "chapter.html"
    course: Coursebuilder
    taskgroups: list['Taskgroupbuilder']


    @property
    def outputfile(self) -> str:
        return f"chapter-{self.name}.html"

    @property
    def sourcefile(self) -> str:
        return f"{self.course.chapterdir}/{self.name}/index.md"

    @property
    def to_be_skipped(self) -> bool:
        return self.skipthis

    @functools.cached_property
    def toc(self) -> str:
        return sdrl.partbuilder.toc(self)

    def as_json(self) -> b.StrAnyDict:
        result = dict(name=self.name, slug=self.name,  # slug for backwards compatibility, TODO 3: remove 2025-01
                      taskgroups=[taskgroup.as_json() for taskgroup in self.taskgroups])
        result.update(super().as_json())
        return result
    
    def process_topmatter(self, sourcefile: str, topmatter: b.StrAnyDict, course: Coursebuilder):
        b.copyattrs(sourcefile,
                    topmatter, self,
                    mustcopy_attrs='title',
                    cancopy_attrs='stage',
                    mustexist_attrs='',
                    overwrite=True)
        self.evaluate_stage(sourcefile, course)

    def _init_from_dict(self, context: str, chapter: b.StrAnyDict):
        super()._init_from_dict(context, chapter)
        self.directory.record_the(Chapter, self.name, self)

    def _init_from_file(self, context: str, course: Coursebuilder):
        self.make_std_dependencies(use_toc_of=self)
        self.find_zipdirs()


class Taskgroup(el.Part):
    TOC_LEVEL = 1  # indent level in table of contents
    course: Course
    chapter: Chapter
    tasks: list['Task']

    def __init__(self, name: str, **kwargs):
        super().__init__(name, **kwargs)
        context = f"taskgroup '{name}' in chapter '{self.chapter.name}'"
        self.course.namespace_add(self)
        self._init_from_dict(context, self.taskgroupdict)
        self._init_from_file(context, self.chapter)
        self._create_tasks()

    def _create_tasks(self):
        self.tasks = [Task(taskdict['name'], parent=self, taskgroup=self).from_json(taskdict)
                      for taskdict in self.taskgroupdict['tasks']]
    
    def _init_from_dict(self, context: str, taskgroupdict: b.StrAnyDict):
        b.copyattrs(context,
                    taskgroupdict, self,
                    mustcopy_attrs='name',
                    cancopy_attrs='tasks, title, slug',
                    mustexist_attrs='taskgroups')

    def _init_from_file(self, context: str, chapter: Chapter):
        pass  # exists only in builder


class Taskgroupbuilder(sdrl.partbuilder.PartbuilderMixin, Taskgroup):
    TEMPLATENAME = "taskgroup.html"
    chapter: Chapterbuilder
    tasks: list['Taskbuilder']

    @property
    def sourcefile(self) -> str:
        return f"{self.course.chapterdir}/{self.chapter.name}/{self.name}/index.md"

    @property
    def to_be_skipped(self) -> bool:
        return self.skipthis or self.chapter.to_be_skipped

    @functools.cached_property
    def toc(self) -> str:
        return sdrl.partbuilder.toc(self)

    def as_json(self) -> b.StrAnyDict:
        result = dict(name=self.name, slug=self.name,  # slug for backwards compatibility, TODO 3: remove 2025-01
                      tasks=[task.as_json() for task in self.tasks])
        result.update(super().as_json())
        return result

    def process_topmatter(self, sourcefile: str, topmatter: b.StrAnyDict, course: Coursebuilder):
        b.copyattrs(sourcefile,
                    topmatter, self,
                    mustcopy_attrs='title',
                    cancopy_attrs='stage',
                    mustexist_attrs='',
                    overwrite=True)
        self.evaluate_stage(sourcefile, course)

    def _add_task(self, task: Taskbuilder):
        task.taskgroup = self
        self.tasks.append(task)

    def _create_tasks(self):
        """Finds and reads task files."""
        self.tasks = []
        filenames = glob.glob(f"{self.course.chapterdir}/{self.chapter.name}/{self.name}/*.md")
        for filename in sorted(filenames):
            if not filename.endswith("index.md"):
                name = os.path.basename(filename[:-3])  # remove .md suffix
                self._add_task(Taskbuilder(name, parent=self, taskgroup=self))

    def _init_from_dict(self, context: str, taskgroupdict: b.StrAnyDict):
        super()._init_from_dict(context, taskgroupdict)
        self.directory.record_the(Taskgroup, self.name, self)

    def _init_from_file(self, context, chapter):
        self.make_std_dependencies(use_toc_of=self)
        self.find_zipdirs()


class MetadataDerivation(el.Step):
    """Copy Topmatter into Parts' attributes, compute assumedby/requiredby/taskorder, check links."""
    course: Coursebuilder

    def do_build(self):
        # ----- copy topmatter into Parts' attributes:
        dir = self.directory
        allparts = list(itertools.chain(dir.get_all(Chapter), dir.get_all(Taskgroup), dir.get_all(Task)))
        for part in allparts:
            topmatter = self.directory.get_the(el.Topmatter, part.name)
            part.process_topmatter(part.sourcefile, topmatter.value, self.course)
            part.evaluate_stage(part.sourcefile, part.course)
        # ----- compute and check stuff:
        self.course.add_inverse_links()
        self.course.compute_taskorder()
        self.course.check_links()



import os

import argparse_subcommand as ap_sub


class SedrilaArgParser(ap_sub.ArgumentParser):
    """One-trick pony class for obtaining the expensive description only when needed."""

    def format_help(self):
        self.description = f"sedrila {self.get_version()}: Tool for 'self-driven lab' (SeDriLa) university courses."
        return super().format_help()

    @staticmethod
    def get_version() -> str:
        import tomllib
        # the development tree (and tar version of the package) have this structure:
        #   pyproject.toml
        #   py/sdrl/argparser.py
        # in contrast, the whl version of the package has this structure:
        #   pyproject.toml
        #   sdrl/argparser.py
        # our logic needs to work for both.  We try the whl version first:
        topdir = os.path.dirname(os.path.dirname(__file__))
        pyprojectfile = os.path.join(topdir, "pyproject.toml")
        if not os.path.exists(pyprojectfile):  # we have the tar or development tree here:
            topdir = os.path.dirname(topdir)  # go from py to top in dev tree
            pyprojectfile = os.path.join(topdir, "pyproject.toml")  # this ought to work now
        with open(pyprojectfile, 'rb') as f:
            toml = tomllib.load(f)
            return toml['tool']['poetry']['version']





"""
Generate the website with incremental build.
See the architecture sketch in README.
TODO:
- 
- use sedrila2.yaml, force titles to be in topmatter
- eventually: kick out author, rename author2 to author, changelog, adjust docs
"""
import argparse
import datetime as dt
import json
import os
import os.path
import typing as tg

import base as b
import cache
import sdrl.constants as c
import sdrl.course
import sdrl.elements as el
import sdrl.directory as dir
import sdrl.macroexpanders as macroexpanders


meaning = """Creates and renders an instance of a SeDriLa course with incremental build.
Checks consistency of the course description.
"""


def add_arguments(subparser: argparse.ArgumentParser):
    subparser.add_argument('--config', metavar="configfile", default=c.AUTHOR_CONFIG_FILENAME,
                           help="SeDriLa configuration description YAML file")
    subparser.add_argument('--include_stage', metavar="stage", default='',
                           help="include parts with this and higher 'stage:' entries in the generated output "
                                "(default: only those with no stage)")
    subparser.add_argument('--log', default="INFO", choices=b.loglevels.keys(),
                           help="Log level for logging to stdout (default: INFO)")
    subparser.add_argument('--sums', action='store_const', const=True, default=False,
                           help="Print task volume reports")
    subparser.add_argument('--clean', action='store_const', const=True, default=False,
                           help="purge cache and perform a complete build")
    subparser.add_argument('targetdir',
                           help=f"Directory to which output will be written.")


def execute(pargs: argparse.Namespace):
    b.set_loglevel(pargs.log)
    targetdir_s = pargs.targetdir
    targetdir_i = _targetdir_i(pargs.targetdir)
    prepare_directories(targetdir_s, targetdir_i)
    create_and_build_course(pargs, targetdir_i, targetdir_s)
    b.finalmessage()


def create_and_build_course(pargs, targetdir_i, targetdir_s) -> sdrl.course.Coursebuilder:
    # ----- prepare build:
    the_cache = cache.SedrilaCache(os.path.join(targetdir_i, c.CACHE_FILENAME), start_clean=pargs.clean)
    b.set_register_files_callback(the_cache.set_file_dirty)
    directory = dir.Directory(the_cache)
    the_course = sdrl.course.Coursebuilder(
        configfile=pargs.config, context=pargs.config, include_stage=pargs.include_stage,
        targetdir_s=targetdir_s, targetdir_i=targetdir_i, directory=directory)
    # ----- perform main part of build:
    prepare_itree_zip(the_course)
    macroexpanders.register_macros(the_course)
    directory.build()
    # ----- build special files:
    b.spit(os.path.join(targetdir_s, c.METADATA_FILE), json.dumps(the_course.as_json(), indent=2))
    generate_htaccess(the_course)
    # ----- clean up and report:
    purge_leftover_outputfiles(directory, targetdir_s, targetdir_i)
    if pargs.sums:
        print_volume_report(the_course)
    the_cache.close()  # write back changes
    return the_course


def generate_htaccess(course: sdrl.course.Coursebuilder):
    if not course.htaccess_template:
        return  # nothing to do
    userlist = [u['webaccount'] for u in course.instructors]
    htaccess_txt = (course.htaccess_template.format(
                       userlist_commas=",".join(userlist),
                       userlist_spaces = " ".join(userlist),
                       userlist_quotes_spaces = " ".join((f'"{u}"' for u in userlist))))
    b.spit(os.path.join(course.targetdir_i, c.HTACCESS_FILE), htaccess_txt)


def prepare_directories(targetdir_s: str, targetdir_i: str):
    # ----- create from scratch if needed:
    if not os.path.exists(targetdir_s):
        os.mkdir(targetdir_s)
    # ----- check plausibility:
    not_empty = len(os.listdir(targetdir_s)) > 1  # targetdir_i will exist even in fresh ones 
    metadatafile = os.path.join(targetdir_s, c.METADATA_FILE)
    if not_empty and not os.path.exists(metadatafile):  # empty pre-existing dirs will fail, too
        b.critical(f"{targetdir_s} does not look like a build directory: {c.METADATA_FILE} is missing")
    # ----- add instructor dir if needed:
    if not os.path.exists(targetdir_i):
        os.mkdir(targetdir_i)


def prepare_itree_zip(the_course: sdrl.course.Coursebuilder):
    if not the_course.itreedir:
        return  # nothing to do
    if not the_course.itreedir.endswith(".zip"):
        b.critical(f"itreedir = '{the_course.itreedir}'; must end with '.zip'")
    the_course.directory.make_the(el.Zipdir, the_course.itreedir, parent=the_course)
    the_course.directory.make_the(el.Zipfile, os.path.basename(the_course.itreedir), parent=the_course, 
                                  sourcefile=the_course.itreedir, instructor_only=True)


def print_volume_report(course: sdrl.course.Coursebuilder):
    """Show total timevalues per stage, difficulty, and chapter."""
    # ----- print cumulative timevalues per stage as comma-separated values (CSV):
    volume_report_per_stage = course.volume_report_per_stage()
    print("date", end="")
    for stage, numtasks, timevalue in volume_report_per_stage.rows:
        print(f",{stage}", end="")
    print("")  # newline
    print(dt.date.today().strftime("%Y-%m-%d"), end="")
    for stage, numtasks, timevalue in volume_report_per_stage.rows:
        print(",%.2f" % timevalue, end="")
    print("")  # newline

    # ----- print all reports as rich tables:
    for report in (volume_report_per_stage,
                   course.volume_report_per_difficulty(),
                   course.volume_report_per_chapter()):
        table = b.Table()
        table.add_column(report.columnheads[0])
        table.add_column(report.columnheads[1], justify="right")
        table.add_column(report.columnheads[2], justify="right")
        totaltasks = totaltime = 0
        for name, numtasks, timevalue in report.rows:
            table.add_row(name,
                          str(numtasks),
                          "%5.1f" % timevalue)
            totaltasks += numtasks
            totaltime += timevalue
        table.add_row("[b]=TOTAL", f"[b]{totaltasks}", "[b]%5.1f" % totaltime)
        b.rich_print(table)  # noqa


def purge_leftover_outputfiles(directory: dir.Directory, targetdir_s: str, targetdir_i: str):
    def keep(outputfile: el.Outputfile) -> bool:
        # two cases: non-Parts, Parts
        return not isinstance(outputfile, el.Part) or not outputfile.to_be_skipped
    
    expected_files = set([of.outputfile for of in directory.get_all_outputfiles() if keep(of)])
    additions_s = {c.AUTHOR_OUTPUT_INSTRUCTORS_DEFAULT_SUBDIR, c.METADATA_FILE}
    additions_i = {c.HTACCESS_FILE}
    purge_all_but(targetdir_s, expected_files | additions_s)
    purge_all_but(targetdir_i, expected_files | additions_i, exception=c.CACHE_FILENAME)


def purge_all_but(dir: str, files: set[str], exception: tg.Optional[str] = None):
    """Delete all files in dir that are not mentioned in files and are no exceptions."""
    actual_files = set(os.listdir(dir))
    delete_these = actual_files - files
    for file in sorted(delete_these):
        if exception and file.startswith(exception):
            continue  # avoid deleting any kind of cache file
        path = os.path.join(dir, file)
        b.info(f"deleted: {path}")
        os.remove(path)


def _targetdir_i(targetdir_s):
    return os.path.join(targetdir_s, c.AUTHOR_OUTPUT_INSTRUCTORS_DEFAULT_SUBDIR)



"""
Markdown rendering with sedrila-specific bells and/or whistles.
"""
import re

import markdown
import markdown.extensions as mde
import markdown.preprocessors as mdpre
import markdown.postprocessors as mdpost

import base as b
import sdrl.macros as macros
import sdrl.replacements as replacements

# see bottom for initialization code


class SedrilaExtension(mde.Extension):
    def extendMarkdown(self, md):
        md.preprocessors.register(SedrilaPreprocessor(md), 'sedrila_preprocessor', 50)
        md.postprocessors.register(SedrilaPostprocessor(md), 'sedrila_postprocessor', 10)


class SedrilaPreprocessor(mdpre.Preprocessor):
    def run(self, lines: list[str]) -> list[str]:
        content = "\n".join(lines)  # we work on the entire markup at once
        content = self.perhaps_suppress_instructorinfo(content)  
        content = self.make_replacements(content)
        content = macros.expand_macros(self.md.context_sourcefile, self.md.partname, content,
                                       is_early_phase=True)
        return content.split("\n")

    def perhaps_suppress_instructorinfo(self, content: str) -> str:
        """in instructor mode, suppress all [INSTRUCTOR::heading] texttexttext [ENDINSTRUCTOR] blocks"""
        if self.md.mode == b.Mode.INSTRUCTOR:
            return content  # leave instructorinfo in instructor mode
        else:               # remove instructorinfo in student mode
            block_re = r"\[INSTRUCTOR::.+?ENDINSTRUCTOR\]"  # non-greedy middle part, just in case
            newcontent = re.sub(block_re, "", content, flags=re.DOTALL)
            nonblock_re = r"\[INSTRUCTOR::.+\]"  # find incomplete blocks that were not removed
            mm = re.search(nonblock_re, newcontent)
            if mm:
                b.error(f"call '{mm.group(0)}' lacks [ENDINSTRUCTOR]", file=md.context_sourcefile)
            return newcontent

    def make_replacements(self, content: str) -> str:
        def the_repl(mm: re.Match) -> str:
            return replacements.get_replacement(self.md.context_sourcefile, mm.group(2), mm.group(1))            
        return re.sub(replacements.replacement_expr_re, the_repl, content, flags=re.DOTALL)


class SedrilaPostprocessor(mdpost.Postprocessor):
    def run(self, text: str) -> str:
        text = macros.expand_macros(self.md.context_sourcefile, self.md.partname, text)
        return text


class SedrilaMarkdown(markdown.Markdown):
    mode: macros.MM
    context_sourcefile: str
    partname: str
    blockmacro_topmatter: dict[str, str]
    includefiles: set[str]  # [INCLUDE::...], [PROT::...] will add a filename here
    termrefs: set[str]  # [TERMREF::...] will add a term alias here


def render_markdown(context_sourcefile: str, partname: str, markdown_markup: str, 
                    mode: b.Mode, blockmacro_topmatter: dict[str, str]) -> b.StrAnyDict:
    """
    Generates HTML from Markdown in sedrila manner.
    See https://python-markdown.github.io/
    """
    md.mode = mode
    md.context_sourcefile = context_sourcefile
    md.partname = partname
    md.blockmacro_topmatter = blockmacro_topmatter
    md.includefiles = set()
    md.termrefs = set()
    html = md.reset().convert(markdown_markup)
    return dict(html=html, includefiles=md.includefiles, termrefs=md.termrefs)

# ######### initialization:

extensions = [SedrilaExtension(), 
              'attr_list', 'codehilite', 'fenced_code',
              'sane_lists', 'tables', 'toc', 'smarty',
              ]
# https://python-markdown.github.io/extensions/attr_list/
# https://python-markdown.github.io/extensions/code_hilite/
# https://python-markdown.github.io/extensions/fenced_code_blocks/
# https://python-markdown.github.io/extensions/sane_lists/
# https://python-markdown.github.io/extensions/smarty/
# https://python-markdown.github.io/extensions/tables/
# https://python-markdown.github.io/extensions/toc/
# https://github.com/daGrevis/mdx_linkify  breaks the correct handling of `a < b` and cannot be used

extension_configs = {
    'toc': {
        # 'slugify':  perhaps replace with numbering-aware version 
    },
    'codehilite': {
        'linenums': True
    },
    'smarty': {
        'smart_quotes': False
    }
}

md = SedrilaMarkdown(extensions=extensions, extension_configs=extension_configs)
# '[TOC]' is Markdown, but looks syntactically like a macro call, so make 'TOC' a macro:
macros.register_macro('TOC', 0, macros.MM.INNER, lambda mc: f"[{mc.macroname}]")



"""
Statistical evaluation (by a course organizer/course developer) of the progress of 
an entire course cohort.

Goals:
- Understand difficulty level distribution across the students.
- Identify tasks with too-low or too-high time value (via worktime) or difficulty (via resubmissions).
- Understand re-submission behavior patterns (per people, per tasks, over time).
- Understand work time patterns (workdays, weekly trends, etc.)
"""
import contextlib
import datetime as dt
import glob
import os

import argparse_subcommand as ap_sub

import base as b
import git
import sdrl.course
import sdrl.participant

meaning = """Statistical evaluation of the progress of an entire course cohort."""


def add_arguments(subparser: ap_sub.ArgumentParser):
    subparser.add_argument('--log', default="INFO", choices=b.loglevels.keys(),
                           help="Log level for logging to stdout (default: INFO)")


class Repo:
    reponame: str


class Commit:
    reponame: str
    timestamp: dt.datetime
    commit_id: str
    by_instructor: bool

class CourseDummy(sdrl.course.Course):
    blockmacro_topmatter = dict()
    
    def __init__(self, *args, **kwargs):
        pass  # we are a truly dumb Dummy!


def execute(pargs: ap_sub.Namespace):
    b.set_loglevel(pargs.log)
    b.info("evaluator, voilà!")
    pull_all_repos()
    collect_commit_metadata()


def pull_all_repos() -> list[str]:
    """Visit subdirs, pull repo (if any), return subdir names with a student.yaml."""
    repolist = []
    # ----- find repolist:
    for path in glob.glob("*"):
        if not os.path.isdir(path):
            continue  # skip files and potential exotic stuff
        if not os.path.isdir(os.path.join(path, '.git')):
            continue  # skip directories not containing a repo
        repolist.append(path)
    repolist.sort()
    # ----- pull each repo:
    b.info(f"pulling {len(repolist)} git repos")
    progressbar = b.get_progressbar(len(repolist))
    for repo in repolist:
        with contextlib.chdir(repo):
            git.pull(silent=True)
            next(progressbar)
    return repolist


def collect_commit_metadata():
    ...


def collect_submisson_yaml_metadata():
    ...




import argparse
import os
import re
import subprocess as sp
import time

import base as b
import git
import sdrl.constants as c
import sdrl.course
import sdrl.interactive as i
import sdrl.participant
import sdrl.repo as r

meaning = """Help instructors evaluate a student's submission of several finished tasks.
"""


def add_arguments(subparser):
    subparser.add_argument('repo_url', action=RepoUrlAction,
                           help="where to find student input")
    subparser.add_argument('--get', default="True", action=argparse.BooleanOptionalAction,
                           help="pull or clone student repo")
    subparser.add_argument('--check', default="True", action=argparse.BooleanOptionalAction,
                           help="perform actual checking via interactive mode or subshell")
    subparser.add_argument('--put', default="True", action=argparse.BooleanOptionalAction,
                           help="commit and push to student repo")
    subparser.add_argument('--interactive', action=argparse.BooleanOptionalAction,
                           help="open interactive terminal interface to approve or reject tasks")
    subparser.add_argument('--override', action='store_true',
                           help="allow overriding of previous faulty gradings")
    subparser.add_argument('--log', default="INFO", choices=b.loglevels.keys(),
                           help="Log level for logging to stdout (default: INFO)")


def execute(pargs: argparse.Namespace):
    b.set_loglevel(pargs.log)
    home_fallback = "."
    if os.environ.get(c.REPOS_HOME_VAR) is None:
        b.warning(f"Environment variable {c.REPOS_HOME_VAR} is not set. "
                  "Assume current directory as student workdirs directory")
        if pargs.repo_url and os.path.isfile(c.PARTICIPANT_FILE):
            b.warning("It looks like you are already inside a student dir. Assuming parent directory instead.")
            home_fallback = ".."
    home = os.environ.get(c.REPOS_HOME_VAR) or home_fallback
    checkout_success = checkout_student_repo(pargs.repo_url, home, pargs.get)
    if not pargs.put and not pargs.check or not checkout_success:
        os.chdir("..")
        return
    student = sdrl.participant.Student()
    course = sdrl.course.CourseSI(configdict=student.metadatadict, context=student.metadata_url)
    commits = git.commits_of_local_repo(reverse=True)
    r.compute_student_work_so_far(course, commits)
    entries, workhours_total, timevalue_total = r.student_work_so_far(course)
    opentasks = rewrite_submission_file(course, c.SUBMISSION_FILE)
    entries = [entry for entry in entries if allow_grading(course, opentasks, entry, pargs.override)]
    if not pargs.check:
        b.info("Don't run check, just prepare commit")
    elif not entries:
        b.info("No tasks to check found. Assuming check already done. Preparing commit.")
    elif pargs.interactive:
        rejections = i.grade_entries(entries, student.course_url, pargs.override)
        if rejections is None:
            b.info("Nothing selected, nothing to do")
            return
        b.spit_yaml(c.SUBMISSION_FILE, r.submission_file_entries(entries, rejections, pargs.override))
        subshell_exit_info(reminder=True)
    else:
        call_instructor_cmd(course, instructor_cmd(), pargs, iteration=0)
    if pargs.put:
        validate_submission_file(course, c.SUBMISSION_FILE)
        commit_and_push(c.SUBMISSION_FILE)
    os.chdir("..")


def allow_grading(course, opentasks, entry, override):
    task = course.task(entry[0])
    if not override:
        requirements = {requirement: course.task(requirement) for requirement in task.requires
                        if course.task(requirement) is not None}  # ignore Taskgroups
        open_requirements = [k for k, v in requirements.items()
                             if not any(k == name for name in opentasks) and not v.is_accepted and v.remaining_attempts]
        if open_requirements:
            return False
    isallowed = (entry[0] in opentasks) != override
    if override:
        return isallowed and (task.rejections or task.is_accepted)
    return (isallowed and (task.remaining_attempts > 0) and
            (not task.is_accepted or opentasks[entry[0]].endswith(c.SUBMISSION_ACCEPT_MARK)))


def checkout_student_repo(repo_url, home, pull=True):
    """Pulls or clones student repo and changes into its directory."""
    inrepo = os.path.isdir(".git") and os.path.isfile("student.yaml")
    if inrepo:
        username = os.path.basename(os.getcwd())
    else:
        username = git.username_from_repo_url(repo_url)
        os.chdir(home)
    if os.path.exists(username) or inrepo:
        if not inrepo:
            os.chdir(username)
        b.info(f"**** pulling repo in existing directory '{os.getcwd()}'")
        if pull:
            if repo_url:
                existing = git.remote_url()
                if repo_url != existing:
                    print(repo_url)
                    print(existing)
                    b.error("user repo with other url already present. please resolve manually. aborting")
                    return False
            git.pull()
        else:
            b.warning("not pulling user repo, relying on existing state")
    else:
        if not pull:
            b.warning("attempted to grade non-existing user without pulling.")
            b.warning("ignore and clone regardless.")
        git.clone(repo_url, username)
        if os.path.exists(username):
            os.chdir(username)
        else:
            b.critical(f"directory '{username}' not found")
        b.info(f"**** cloned repo into new directory '{os.getcwd()}'")
    return True


def read_submission_file() -> dict[str, str]:
    return b.slurp_yaml(c.SUBMISSION_FILE)


def rewrite_submission_file(course: sdrl.course.Course, filename: str):
    """Checks status of entries and inserts different marks where needed."""
    entries = b.slurp_yaml(filename)
    rewrite_submission_entries(course, entries)
    b.spit_yaml(filename, entries)
    return {taskname: mark for taskname, mark in entries.items() if mark != c.SUBMISSION_NONTASK_MARK}


def rewrite_submission_entries(course: sdrl.course.Course, entries: b.StrAnyDict):
    """Checks status of entries and inserts different marks where needed."""
    for taskname, mark in entries.items():
        task = course.task(taskname)
        if task is None:
            entries[taskname] = c.SUBMISSION_NONTASK_MARK
        elif task.is_accepted:
            entries[taskname] = c.SUBMISSION_ACCEPT_MARK
        elif task.rejections > 0:
            entries[taskname] += f" (previously rejected {task.rejections}x)"


def subshell_exit_info(reminder: bool = False):
    if reminder and not(os.environ.get(c.SEDRILA_COMMAND_ENV)):
        return  # explicit call, no subshell. no need to provide that info
    if reminder:
        b.info("You are still inside a subshell command!")
    b.info("Exit that command (e.g. the shell) to trigger automatic commit+push")
    b.info(f"of the modified {c.SUBMISSION_FILE}.")


def call_instructor_cmd(course: sdrl.course.Course, cmd: str,  # noqa
                        pargs: argparse.Namespace = None, iteration: int = 0):
    """Calls user-set command as indicated by environment variables"""
    if iteration == 0:
        b.info(f"Will now call the command given in the {c.REPO_USER_CMD_VAR} environment variable or else")
        b.info(f"the command given in the SHELL environment variable or else '{c.REPO_USER_CMD_DEFAULT}'.")
        b.info(f"The resulting command in your case will be:\n  '{cmd}'")
        subshell_exit_info()
        b.info(f"Please change status of tasks in {c.SUBMISSION_FILE} to either {c.SUBMISSION_ACCEPT_MARK}\n"
               f"or {c.SUBMISSION_REJECT_MARK} accordingly.")
        b.info("You can also just run `sedrila` to get an interactive list.")
        b.info("Feel free to add remarks at the end of the accept/reject lines.")
    else:
        b.info(f"Calling '{cmd}' again. (You can Ctrl-C right after it.)")
    if pargs:
        os.environ[c.SEDRILA_COMMAND_ENV] = f"instructor {pargs.repo_url} --interactive --no-get --no-put"
    sp.run(cmd, shell=True)
    time.sleep(0.8)  # give user a chance to hit Ctrl-C
    if pargs:
        del os.environ[c.SEDRILA_COMMAND_ENV]


def instructor_cmd() -> str:
    return os.environ.get(c.REPO_USER_CMD_VAR) or os.environ.get('SHELL', c.REPO_USER_CMD_DEFAULT)


def validate_submission_file(course: sdrl.course.Course, filename: str) -> bool:
    """Check whether the submission file contains (and only contains) sensible entries."""
    entries = b.slurp_yaml(filename)
    has_accept = any((r.is_accepted(mark) for taskname, mark in entries.items()))
    has_reject = any((r.is_rejected(mark) for taskname, mark in entries.items()))
    allowable_marks = (f"(?:{c.SUBMISSION_OVERRIDE_PREFIX})?{c.SUBMISSION_ACCEPT_MARK}|"
                       f"(?:{c.SUBMISSION_OVERRIDE_PREFIX})?{c.SUBMISSION_REJECT_MARK}|{c.SUBMISSION_CHECK_MARK}")
    is_valid = True
    
    def error(msg: str) -> bool:
        b.error(msg)
        return False
    if not has_accept and not has_reject:
        is_valid = b.error(f"Invalid {filename}: has neither {c.SUBMISSION_ACCEPT_MARK} nor "
                           f"{c.SUBMISSION_REJECT_MARK} marks.")
    for taskname, mark in entries.items():
        if not course.task(taskname):
            is_valid = error(f"No such task exists: {taskname}")
        if not re.match(allowable_marks, mark):
            is_valid = error(f"Impossible mark: \"{taskname}: {mark}\"")
    return is_valid


def commit_and_push(filename: str):
    assert filename == c.SUBMISSION_FILE  # our only purpose here, the arg is for clarity
    git.commit(*[filename], msg=f"{c.SUBMISSION_FILE} checked", signed=True)
    git.push()


class RepoUrlAction(argparse.Action):  # allow repo_url to be optional without --
    def __init__(self, option_strings, dest, nargs=None, **kwargs):  # noqa
        kwargs['required'] = not(os.path.isfile(c.PARTICIPANT_FILE))
        super().__init__(option_strings, dest, **kwargs)

    def __call__(self, parser, namespace, values, option_string=None):
        setattr(namespace, self.dest, values)



"""Technical operations for reading information from git repos."""
import dataclasses
import datetime as dt
import getpass
import os
import re
import subprocess as sp
import typing as tg

import base as b

LOG_FORMAT_SEPARATOR = '\t'
LOG_FORMAT = "%h%x09%ae%x09%at%x09%GF%x09%s"  # see notes at attributes of Commit or git help log "Pretty Formats"

@dataclasses.dataclass
class Commit:
    hash: str  # %h
    author_email: str  # %ae
    author_date: dt.datetime  # converted from %at
    key_fingerprint: str  # %GF
    subject: str  # %s
    files: int
    insertions: int
    deletions: int


def add(filename: str):
    os.system(f"git add {filename}")    


def clone(repo_url: str, targetdir: str):
    os.system(f"git clone {repo_url} {targetdir}")


def commit(*filenames, msg, **kwargs):
    for filename in filenames:
        add(filename)
    signed = kwargs.pop('signed', False)
    os.system(f"git commit {'-S ' if signed else ''}-m'{msg}'")


def commits_of_local_repo(reverse=False, with_insertions_deletions=False) -> tg.Sequence[Commit]:
    """Returns all commits in youngest-first order (like git log), or reversed."""
    if with_insertions_deletions:
        return commits_of_local_repo_with_insertions_deletions(reverse)
    gitcmd = ["git", "log", f"--format=format:{LOG_FORMAT}"]
    gitrun = sp.run(gitcmd, capture_output=True, encoding='utf8', text=True)
    result = []
    for line in gitrun.stdout.split('\n'):
        hash_, email, tstamp, fngrprnt, subj = tuple(line.split(LOG_FORMAT_SEPARATOR))
        c = Commit(hash_, email, 
                   dt.datetime.fromtimestamp(int(tstamp), tz=dt.timezone.utc),
                   fngrprnt, subj, 0, 0, 0)
        result.append(c)
    return list(reversed(result)) if reverse else result


def commits_of_local_repo_with_insertions_deletions(reverse=False):
    gitcmd = ["git", "log", f"--format=format:{LOG_FORMAT}", "--shortstat"]
    # produces a 3-line block per commit:
    # d0ac3d8 author@domain       1725618158              Testabgrenzung.md: feedback to question the need for this task again at the end
    #  1 file changed, 2 insertions(+), 1 deletion(-)
    # 
    # where line 1 is TAB-delimited, line2 is text in varying formats, and line3 is empty.
    gitrun = sp.run(gitcmd, capture_output=True, encoding='utf8', text=True)
    result = []
    for block in gitrun.stdout.split('\n\n'):
        print("#####", block, block.split('\n'))
        lines = block.split('\n')  # infoline, statline, emptyline
        hash_, email, tstamp, fngrprnt, subj = tuple(lines[0].split(LOG_FORMAT_SEPARATOR))
        mm = re.search(r'((?P<files>\d+) file.+)?((?P<insertions>\d+) insertion.+)?((?P<deletions>\d+) deletion)?',
                       lines[1])
        if mm:
            files = int(mm.group("files")) if mm.group("files") else 0
            insertions = int(mm.group("insertions")) if mm.group("insertions") else 0
            deletions = int(mm.group("deletions")) if mm.group("deletions") else 0
        else:
            files = insertions = deletions = 0
        c = Commit(hash_, email, 
                   dt.datetime.fromtimestamp(int(tstamp), tz=dt.timezone.utc),
                   fngrprnt, subj,
                   files, insertions, deletions)
        result.append(c)
    return list(reversed(result)) if reverse else result


def contents_of_file_version(refid: str, filename: str, encoding=None) -> tg.AnyStr:
    raw = sp.check_output(f"git show {refid}:{filename}", shell=True)
    if encoding:
        return raw.decode(encoding=encoding)
    else:
        return raw


def discard_commits(howmany: int):
    os.system(f"git reset --hard HEAD~{howmany}")


def origin_remote_of_local_repo() -> str:
    """The local repo's 'origin' remote, which we assume to exist and to be the relevant remote."""
    git_remote = sp.check_output("git remote -v show", shell=True).decode('utf8')
    # e.g.:  origin  git@github.com:myaccount/myrepo.git (fetch)
    fetchremote_regexp = r"origin\s+(\S+)\s+\(fetch\)"
    mm = re.search(fetchremote_regexp, git_remote)
    if not mm:
        b.critical(f"cannot find 'origin' URL in  git remote  output:\n{git_remote}")
    return mm.group(1)


def pull(silent=False):
    if silent:
        try:
            pull_output = sp.check_output("git pull --ff-only", 
                                          stderr=sp.STDOUT, shell=True).decode('utf8')
        except sp.CalledProcessError as err:
            print(err.stdout)
    else:
        os.system("git pull --ff-only")


def push():
    os.system("git push")


def remote_url():
    result = sp.run(["git", "config", "--get", "remote.origin.url"], stdout=sp.PIPE)
    return result.stdout.decode("utf-8").strip()

    
def username_from_repo_url(repo_url: str) -> str:
    # a repo_url is git@server:useraccount/reponame.git or git@server:useraccount/subset/reponame.git
    repo_url_regexp = r":([\w_\.-]+)/"
    mm = re.search(repo_url_regexp, repo_url)
    if repo_url.startswith("http") or not mm:
        #for testing purposes, local paths are allowed but default to the current user
        if os.path.isdir(repo_url):
            return getpass.getuser()
        if not mm:
            repo_url_regexp = r"/([\w_\.-]+)/([\w_\.-]+).git"
            mm = re.search(repo_url_regexp, repo_url)
        if mm and mm.group(1):
            return mm.group(1)
        b.critical(f"Git url '{repo_url}' is not usable.\nNeed one like 'git@server:useraccount/reponame.git'.")
    return mm.group(1) if mm else None  # tests will patch b.critical away
